@@ -1,6 +1,6 @@
 package Gtk2::Ex::HyperTextBuffer;
 
-our $VERSION = '0.17';
+our $VERSION = '0.19';
 
 use strict;
 use Gtk2;
@@ -74,12 +74,15 @@ Simple constructor.
 sub INIT_INSTANCE {
 	my $self = shift;
 	$self->{edit_mode_tags} = [];
+	$self->{undo} = [];
+	$self->{redo} = [];
 	$self->signal_connect(mark_set => sub {
 			my ($self, $iter, $mark) = @_;
 			$self->set_edit_mode_from_cursor()
 				if $mark->get_name eq 'insert';
 		} );
 	$self->signal_connect_after(insert_text => \&on_insert);
+	$self->signal_connect(delete_range => \&on_delete_range);
 }
 
 =item C<clear()>
@@ -116,7 +119,7 @@ the format as used by L<Pod::Simple> but it uses more verbose element names.
 
 For example:
 
-	['Document', {},
+	['Page', {},
 		['head1', {}, 'Test page'],
 		"\n\n",
 		"This is a test page showing some syntax:\n\n",
@@ -127,19 +130,25 @@ For example:
 		], "\n",
 	]
 
+This function creates a new unmodified buffer.
+This action is not un-doable.
+
 =cut
 
 sub set_parse_tree {
 	my ($self, $tree) = @_;
 	#use Data::Dumper; print "HyperTextBuffer got tree: ", Dumper $tree;
 	$self->clear;
+	$self->{_undo_lock} = 1;
 	$self->insert_blocks_at_cursor($tree);
+	$self->{_undo_lock} = 0;
 	$self->set_modified(0);
 }
 
 =item C<insert_blocks(ITER, BLOCK, ...)>
 
 Insert one or more partial parse trees in the buffer at ITER.
+The blocks will also be added to the undo stack.
 
 =cut
 
@@ -154,6 +163,7 @@ sub insert_blocks {
 	$self->place_cursor($iter);
 	$self->insert_blocks_at_cursor(@_);
 	$self->place_cursor( $self->get_iter_at_mark($mark) );
+	$self->delete_mark($mark);
 }
 
 =item C<insert_blocks_at_cursor(BLOCK, ...)>
@@ -170,7 +180,10 @@ sub insert_blocks_at_cursor {
 	my $self = shift;
 	
 	$self->signal_handlers_block_by_func(\&on_insert);
+	my $begin = $self->get_iter_at_mark( $self->get_insert )->get_offset;
 	$self->_insert_blocks(@_);
+	my $end = $self->get_iter_at_mark( $self->get_insert )->get_offset;
+	$self->_record_step('I', $begin, $end, @_);
 	$self->signal_handlers_unblock_by_func(\&on_insert);
 	return 1;
 }
@@ -180,7 +193,6 @@ sub _insert_blocks {
 
 	for (@blocks) {
 		unless (ref $_) {
-			$_ =~ s/^(\s*)\*(\s+)/$1\x{2022}$2/mg unless $_Verbatim; # FIXME doesn't belong in this module !
 			if (@{$self->{edit_mode_tags}}) {
 				my $iter = $self->get_iter_at_mark( $self->get_insert );
 				$self->insert_with_tags(
@@ -192,7 +204,7 @@ sub _insert_blocks {
 			my ($type, $meta) = @$_[0, 1];
 			my $tag;
 			if ($type eq 'image') {
-				$self->insert_image_from_file($$meta{src});
+				$self->insert_image_from_file(@$meta{qw/src width height/});
 				next;
 			}
 			elsif ($type eq 'link') {
@@ -208,7 +220,7 @@ sub _insert_blocks {
 				pop @{$self->{edit_mode_tags}};
 				$_Verbatim-- if $type eq 'Verbatim'; # FIXME
 			}
-			elsif ($type eq 'Document' or $type eq 'Para') {
+			elsif ($type eq 'Page' or $type eq 'Para') {
 				if ($$meta{base}) {
 					$self->{base} = $$meta{base};
 					$self->{base} =~ s#^file://##;
@@ -226,26 +238,43 @@ sub _insert_blocks {
 
 }
 
-=item C<insert_image_from_file(FILENAME, ITER)>
+=item C<insert_image_from_file(FILENAME, WIDTH, HEIGHT, ITER)>
 
 Insert an image in the buffer at the cursor
 or at ITER if ITER is specified.
 
+WIDTH and HEIGHT are optional arguments. If both are set the image
+is resized to these dimensions. If one of both is zero or undefined
+the image is scaled with fixed ratio.
+
+Use C<insert_blocks()> if you want this action to be un-doable. 
+
 =cut
 
 sub insert_image_from_file {
-	my ($self, $src, $iter) = @_;
+	my ($self, $src, $w, $h, $iter) = @_;
+	($w, $iter) = (undef, $w) if @_ == 3 and ref $w;
 	my $file = $self->{base} ? File::Spec->rel2abs($src, $self->{base}) : $src;
 	#warn "file: $src, base: $self->{base} => $file\n";
 
 	my $pixbuf = undef;
-	eval { $pixbuf = Gtk2::Gdk::Pixbuf->new_from_file($file) }
-		if (-f $file and -r _) ;
+	if (-f $file and -r _) {
+		unless ($w || $h) {
+			eval { $pixbuf =
+				Gtk2::Gdk::Pixbuf->new_from_file($file) }
+		}
+		else {
+			eval { $pixbuf =
+				Gtk2::Gdk::Pixbuf->new_from_file_at_size(
+					$file, $w || -1, $h || -1) }
+		}
+	}
 	$pixbuf ||= Gtk2::Image->new->render_icon('gtk-missing-image', 'dialog')->copy;
 	die "Could not insert image: $file\n" unless $pixbuf;
 	
 	$pixbuf->{image_src} = $src;
 	$pixbuf->{image_file} = $file;
+	$pixbuf->{resized} = ($w || $h) ? [$w, $h] : 0;
 	$iter ||= $self->get_iter_at_mark( $self->get_insert );
 
 	#my $mark = $self->create_mark(undef, $iter, 0);
@@ -253,7 +282,7 @@ sub insert_image_from_file {
 	
 	$self->insert_pixbuf($iter, $pixbuf);
 
-	return 1;
+	return $pixbuf;
 }
 
 =item C<get_parse_tree()>
@@ -289,7 +318,6 @@ sub get_parse_tree {
 		
 		EXTRACT_TAGS:
 		my $slice = $previous->get_slice($iter);
-		$slice =~ s/^(\s*)\x{2022}(\s+)/$1\*$2/mg; # FIXME totally wrong level
 		if ($slice =~ /\x{fffc}/) {
 			my $image = $previous->copy;
 			while ($slice =~ s/^(.*?)\x{fffc}//s) { # Embeded image
@@ -297,8 +325,10 @@ sub get_parse_tree {
 				$image->forward_chars(length $1);
 				my $pixbuf = $image->get_pixbuf;
 				if ($pixbuf) {
-					my $src = $pixbuf->{image_src};
-					push @{$stack[-1]}, ['image', {src => $src}];
+					my $meta = { src => $pixbuf->{image_src} };
+					@$meta{qw/width height/} = @{$pixbuf->{resized}}
+						if $pixbuf->{resized};
+					push @{$stack[-1]}, ['image', $meta];
 				}
 				$image->forward_char;
 			}
@@ -374,7 +404,7 @@ sub get_parse_tree {
 			: length($_)
 	} @$root;
 		
-	unshift @$root, 'Document', {};
+	unshift @$root, 'Page', {};
 	#use Data::Dumper; warn "get_parse_tree: ", Dumper $root;
 	return $root;
 }
@@ -509,17 +539,6 @@ sub select_link {
 	}
 }
 
-=item C<apply_tag_to_selection>
-
-TODO
-
-=cut
-
-sub apply_tag_to_selection {
-	my ($self, $tag) = @_;
-	
-}
-
 =item C<replace_selection(TEXT)>
 
 Replace any selected text with TEXT.
@@ -628,24 +647,74 @@ sub on_insert {
 	$start->backward_chars(length $string);
 	$self->remove_all_tags($start, $end);
 	$self->apply_tag($_, $start, $end) for @{$self->{edit_mode_tags}};
+	$self->_record_step('I', $start->get_offset, $end->get_offset, $string);
+}
+
+sub on_delete_range {
+	my ($self, $begin, $end) = @_;
+	my $tree = $self->get_parse_tree($begin, $end);
+	$self->_record_step('D', $begin->get_offset, $end->get_offset, $tree);
 }
 
 =item C<undo>
 
 TODO
 
-=cut
-
-sub undo {
-}
-
 =item C<redo>
 
 TODO
 
 =cut
 
+my %reverse_step = (
+	I => 'D',
+	D => 'I',
+	A => 'R',
+	R => 'A',
+);
+
+sub undo {
+	my $self = shift;
+	return 0 unless @{$self->{undo}};
+	my $step = pop @{$self->{undo}};
+	push @{$self->{redo}}, [@$step];
+	$$step[0] = $reverse_step{$$step[0]};
+	$self->_do_step(@$step);
+	return 1;
+}
+
 sub redo {
+	my $self = shift;
+	return 0 unless @{$self->{redo}};
+	my $step = pop @{$self->{redo}};
+	push @{$self->{undo}}, [@$step];
+	$self->_do_step(@$step);
+	return 1;
+}
+
+sub _record_step {
+	my ($self, $step, $begin, $end, @data) = @_;
+	return if $self->{_undo_lock};
+	$self->{_redo} = [];
+	push @{$self->{undo}}, [$step, $begin, $end, @data];
+}
+
+sub _do_step {
+	my ($self, $step, $begin, $end, @data) = @_;
+	$self->{_undo_lock} = 1;
+	$_ = $self->get_iter_at_offset($_) for $begin, $end;
+	if ($step eq 'I') { # insert
+		$self->insert_blocks($begin, @data);
+	}
+	elsif ($step eq 'D') { # delete
+		$self->delete($begin, $end);
+	}
+	elsif ($step eq 'A') { # add tag
+	}
+	elsif ($step eq 'R') { # remove tag
+	}
+	else { die "BUG: unknown step type: $step" }
+	$self->{_undo_lock} = 0;
 }
 
 =item C<copy_clipboard>
@@ -761,39 +830,34 @@ sub paste_clipboard {
 		replace     => $replace,
 	};
 	
-	$clipboard->request_contents(
-		Gtk2::Gdk::Atom->intern('ZIM_TREE', 0),
-		\&_clipboard_paste, $paste_data   );
-}
-
-sub _clipboard_paste {
-	# Paste data type negotiating
-	my ($clipboard, $selection, $paste_data) = @_;
-	my $type = $selection->target->name;
-	my $data = $selection->data;
-	#print STDERR '_clipboard_paste type: '.$selection->target->name.", data: $data\n";
-	
-	if ($type eq 'ZIM_TREE' and $_clipboard and $data eq "$_clipboard") {
-		$paste_data->{type} = $type;
-		&_clipboard_paste_data($clipboard, $data, $paste_data);
+	my %targets = map {($_->name => $_)} $clipboard->wait_for_targets();
+	#warn "Available targets: ".join(' ', keys %targets)."\n";
+	my ($atom) = grep defined($_),
+		@targets{qw#ZIM_TREE text/x-zim-page-list text/uri-list#};
+	if ($atom) {
+		$clipboard->request_contents(
+			$atom, \&_clipboard_paste, $paste_data);
 	}
 	else {
-		# Not pasting from a zim widget 
-		# or pasting from a different process
-		$paste_data->{type} = 'STRING';
-		$clipboard->request_text(\&_clipboard_paste_data, $paste_data);
+		$clipboard->request_text(
+			\&_clipboard_paste, $paste_data);
 	}
 }
 
-sub _clipboard_paste_data {
-	# Actual paste
-	my ($clipboard, $string, $paste_data) = @_;
+sub _clipboard_paste { # callback which does the actual paste
+	my ($clipboard, $data, $paste_data) = @_;
 	my $self = $paste_data->{buffer};
-	my $type = $paste_data->{type};
-	#print STDERR "_clipboard_paste_data type: $type, string: $string\n";
+	my $type = ref($data) ? $data->target->name : 'text/plain' ;
+	my $string = ref($data) ? $data->data : $data;
 
 	return unless length $string;
-	return if $type eq 'ZIM_TREE' and "$string" ne "$_clipboard";
+	if ($type eq 'ZIM_TREE' and "$string" ne "$_clipboard") {
+		# not same process - fall back
+		$clipboard->request_text(
+			\&_clipboard_paste, $paste_data);
+		return;
+	}
+	#warn "pasting: $type\n$string\n...\n";
 
 	$self->begin_user_action if $paste_data->{interactive};
 	
@@ -808,15 +872,14 @@ sub _clipboard_paste_data {
 		$iter = $self->get_iter_at_mark( $mark );
 		$self->delete_mark($mark);
 	}
-	else { $iter = $self->get_iter_at_mark( $self->get_insert ) }
 	
 	# actual insert
 	my $tree;
-	if ($paste_data->{type} eq 'ZIM_TREE') { $tree = $_clipboard }
-	else { # text
+	if ($type eq 'ZIM_TREE') { $tree = $_clipboard }
+	elsif ($type eq 'text/plain') {
 		# FIXME url parsing copy-pasted from Zim/Formats/Wiki.pm
 		my $i = 0;
-		my @tree = ('Document', {},
+		my @tree = ('Page', {},
 			map   { ($i++ % 2) ? _parse_link($_) : $_ }
 			split m#(
 				\b\w[\w\+\-\.]+://  \S*\[\S+\](?:\S+[\w\/])? |
@@ -830,8 +893,18 @@ sub _clipboard_paste_data {
 			# See rfc3986 for the official -but unpractical- regex
 		$tree = \@tree;
 	}
+	else { # text/uri-list or text/x-zim-page-list
+		my @links = Zim::GUI::Component->decode_uri_list($string);
+		if ($type eq 'text/x-zim-page-list') {
+			my $rep = $self->{_zim_name}; # FIXME FIXME FIXME
+			@links = map {s/^\Q$rep\E\?//; $_} @links;
+		}
+		$tree = [ 'Page', {}, 
+			map {(['link', {to => $_}, $_], "\n")} @links ];
+	}
 
-	$self->insert_blocks($iter, $tree);
+	if ($iter) { $self->insert_blocks($iter, $tree)           }
+	else       { $self->insert_blocks_at_cursor($iter, $tree) }
 	$self->end_user_action if $paste_data->{interactive};
 }
 
