@@ -3,8 +3,10 @@ package Zim::Formats::Html;
 use strict;
 use Encode;
 use Zim::Template;
+use Zim::Formats;
 
-our $VERSION = '0.17';
+our $VERSION = '0.18';
+our @ISA = qw/Zim::Formats/;
 
 =head1 NAME
 
@@ -30,10 +32,11 @@ our %entities = (
 	'>' => 'gt',
 );
 	
-sub save_tree {
-	my ($class, $io, $tree, $page) = @_;
+sub save_tree { # HACK: template and temp_data are unsupported args
+	my ($class, $io, $tree, $page, $template, $temp_data) = @_;
+	#use Data::Dumper; warn Dumper $tree;
 
-	my $template = $page->{repository}{_template};
+	$template ||= $page->{repository}{_template} || '';
 	$template = length($template)
 		? Zim::Template->new($template)
 		: Zim::Template->new(\'[% page.body %]')
@@ -60,12 +63,12 @@ sub save_tree {
 			title => $title,
 			heading => $heading,
 			links => sub {
-				[ map { {name => $_, file => _href($page, $_)} }
-					sort $page->_list_links ]
+				[ map { {name => $_, file => href($page, $_)} }
+					sort $page->list_links ]
 			},
 			backlinks => sub {
-				return [ map { {name => $_, file => _href($page, $_)} }
-					sort $page->_list_backlinks ] ;
+				return [ map { {name => $_, file => href($page, $_)} }
+					sort $page->list_backlinks ] ;
 			},
 		},
 		prev => {},
@@ -80,7 +83,7 @@ sub save_tree {
 		eval {
 			$page->{_mode} = lc $data->{mode};
 			print qq#<div class="slide">\n# if $page->{_mode} eq 's5';
-			_dump_node($tree, $page);
+			$class->_dump_node($tree, $page);
 			print qq#</div>\n# if $page->{_mode} eq 's5';
 		};
 		select $old_fh;
@@ -91,21 +94,23 @@ sub save_tree {
 		my $prev = $page->{repository}{_prev}->name;
 		$data->{prev} = {
 			name => $prev,
-			file => _href($page, $prev),
+			file => href($page, $prev),
 		};
 	}
 	if ($page->{repository}{_next}) {
 		my $next = $page->{repository}{_next};
 		$data->{next} = {
 			name => $next,
-			file => _href($page, $next),
+			file => href($page, $next),
 		};
 	}
+	$data = {%$data, %$temp_data} if $temp_data;
 
 	$template->process($data => $io);
 }
 
 sub _dump_node {
+	my $class = shift;
 	no warnings;
 	
 	my ($node, $page) = @_;
@@ -115,9 +120,9 @@ sub _dump_node {
 	if ($name =~ /^head(\d+)/) { @tags = ("<h$1>", "</h$1>\n") }
 	elsif ($name eq 'Para') {
 		@tags = ("<p>\n", "\n</p>\n");
-		if (! ref $$node[0] and $$node[0] =~ /^\s*\*\s+/m) {
+		if (grep {!ref($_) and /^\s*\x{2022}\s+/m} @$node) {
 			# ugly hack ... should not be necessary
-			parse_list($page, @$node);
+			$class->parse_list($page, @$node);
 			@$node = ();
 		}
 		elsif ( $page->{_mode} eq 's5'
@@ -137,20 +142,25 @@ sub _dump_node {
 	elsif ($name eq 'link') {
 		my ($type, $href) = $page->parse_link($$opt{to});
 		if ($type eq 'page') {
-			my $link = $page->resolve_link($$opt{to});
-			$href = _href($page, $link->name) if $link;
+			$href = href($page, $page->resolve_name($href));
 		}
-		print STDERR "  link: $$opt{to} => $href\n";
+		elsif ($type eq 'file') {
+			$href = href_file($href);
+		}
+		warn "#  link: $$opt{to} => $href\n";
 		@tags = ("<a href='$href'>", "</a>");
 	}
 	elsif ($name eq 'image') {
-		@tags = ("<img src='$$opt{src}' alt='$$opt{src}' />", '');
+		my $src = href_file($$opt{src});
+		my $alt = $src;
+		$alt =~ s#.*[/\\]##g;
+		@tags = ("<img src='$src' alt='$name' />", '');
 	}
 	# else recurs silently
 
 	print $tags[0];
 	for (@$node) {
-		if (ref $_) { _dump_node($_, $page) } # recurse
+		if (ref $_) { $class->_dump_node($_, $page) } # recurse
 		else {
 			s#([<>&])#\&$entities{$1}\;#g; # encode
 			s#\n#<br />\n#g unless $tags[0] =~ m#<pre>#;
@@ -160,13 +170,33 @@ sub _dump_node {
 	print $tags[1];
 }
 
-sub _href {
+=item C<href(PAGE, LINK)>
+
+Function that returns an url for a link.
+
+=item C<href_file(FILE)>
+
+Function that returns an url for a file link.
+
+=cut
+
+sub href_file {
+	my $href = shift;
+	$href = Encode::encode_utf8($href); # turn utf8 flag off
+	$href =~ s{ ([^A-Za-z0-9\-\_\.\!\~\*\'\(\)\/\:]) }
+	          { sprintf("%%%02X",ord($1))            }egx;
+	# url encoding - char set from man uri(7), see relevant rfc
+	# added '/' and ':' to char set for readability of uris
+	return $href;
+}
+
+sub href {
 	my ($page, $href) = @_;
 	return '' unless length $href;
 	my $namespace = $page->namespace;
 	#print STDERR "page: $href, relative: $namespace => ";
 	$href =~ s#^:*#:#;
-	unless ($href =~ s#^($namespace):*##) {
+	unless ($href =~ s#^(\Q$namespace\E):*##) {
 		$href =~ s/^:*//;
 		$href = $page->{_root_dir}.$href;
 	}
@@ -175,24 +205,27 @@ sub _href {
 	$href .= '.html';
 	#print STDERR $href."\n";
 	$href = Encode::encode_utf8($href); # turn utf8 flag off
-	$href =~ s#([^a-zA-Z0-9/\.])#sprintf '%%%02x', ord $1#eg; # url encoding
+	$href =~ s{ ([^A-Za-z0-9\-\_\.\!\~\*\'\(\)\/\:]) }
+	          { sprintf("%%%02X",ord($1))            }egx;
+	# url encoding - char set from man uri(7), see relevant rfc
+	# added '/' and ':' to char set for readability of uris
 	return $href;
 }
 
-sub parse_list { # FIXME totally wrong level for this kind of parsing :(
+sub parse_list { # FIXME wrong level for this kind of parsing :(
+	my $class = shift;
 	no warnings;
 
 	my ($page, @nodes) = @_;
 	
-	print "<ul>\n";
-	my ($level, $close);
+	my ($level, $close) = (-1, '');
 	for (@nodes) {
-		if (ref $_) { _dump_node($_, $page) }
+		if (ref $_) { $class->_dump_node($_, $page) }
 		else {
 			s#([<>&])#\&$entities{$1}\;#g; # encode
 			my @lines = split /[\n\r]+/, $_;
 			for (@lines) {
-				if ($_ =~ s/^(\s*)\*\s+//) {
+				if ($_ =~ s/^(\s*)\x{2022}\s+//) {
 					print $close;
 					my $lvl = length $1;
 					if ($lvl > $level) {
