@@ -18,7 +18,7 @@ use Zim::File;
 eval "use Gtk2::Ex::DesktopEntryMenu";
 my $has_mimeapplications = $@ ? 0 : 1;
 
-our $VERSION = '0.17';
+our $VERSION = '0.18';
 
 our @ISA = qw/Zim::GUI::Component/;
 
@@ -35,8 +35,7 @@ Zim::GUI::PageView - Page TextView widgets
 
 This module contains the widgets to display an editable
 text buffer containing the current page. It includes a search entry
-at the bottom of the TextView, formatting codes for the TextBuffer and 
-an undo stack.
+at the bottom of the TextView, and autoformatting logic.
 
 =head1 METHODS
 
@@ -46,9 +45,6 @@ Undefined methods are AUTOLOADED to the Gtk2::Ex::HyperTextView object.
 
 =cut
 
-our $LINK_ICON = Gtk2->CHECK_VERSION(2, 6, 0) ? 'gtk-connect' : 'gtk-convert';
-	# gtk-connect stock item was introduced in 2.6.0
-
 my $ui_format_actions =
 # name,		stock id,	label,		accelerator,	tooltip
 qq{
@@ -67,7 +63,7 @@ Verbatim	.		_Verbatim	<ctrl>T		Verbatim
 
 my $ui_format_toggle_actions =
 qq{
-TLink		$LINK_ICON	_Link		.		Link
+TLink		gtk-connect	_Link		.		Link
 TBold		gtk-bold	_Bold		.		Bold
 TItalic		gtk-italic	_Italic		.		Italic
 TUnderline	gtk-underline	_Underline	.		Underline
@@ -76,14 +72,12 @@ TStrike		gtk-strikethrough	Stri_ke	.		Strike
 
 my $ui_actions =
 qq{
-SplitPage	.		Spli_t page	.		Split page
-MergePage	.		_Merge page	.		Merge page
 Undo		gtk-undo	_Undo		<ctrl>Z		Undo
 Redo		gtk-redo	_Redo		<ctrl><shift>Z	Redo
 Delete		gtk-delete	_Delete		.		Delete
 EditLink	gtk-properties	_Edit Link...	<ctrl>E		Edit link
 InsertDate	.		Insert _Date and Time	<ctrl>D	Insert date
-InsertLink	$LINK_ICON	Insert _Link...		.	Insert link
+InsertLink	gtk-connect	Insert _Link...		.	Insert link
 InsertImage	.		Insert _Image...	.	Insert image
 FindReplace	gtk-find-and-replace	_Replace...	<ctrl>H	Find and Replace
 };
@@ -97,7 +91,7 @@ Paste		gtk-paste	_Paste		<ctrl>V		Paste
 
 my $ui_actions_ro =
 qq{
-Link		$LINK_ICON	_Link		<ctrl>L		Link
+Link		gtk-connect	_Link		<ctrl>L		Link
 Find		gtk-find	_Find...	<ctrl>F		Find
 FindNext	.		Find Ne_xt	<ctrl>G		Find next
 FindPrev	.		Find Pre_vious	<ctrl><shift>G	Find previous
@@ -190,17 +184,10 @@ Simple constructor.
 
 =cut
 
-my ($k_tab, $k_return, $k_kp_enter, $k_backspace, $k_escape, $k_multiply, $k_home, $k_F3) =
-	@Gtk2::Gdk::Keysyms{qw/Tab Return KP_Enter BackSpace Escape KP_Multiply Home F3/};
+my ($k_tab, $k_l_tab, $k_return, $k_kp_enter, $k_backspace, $k_escape, $k_multiply, $k_home, $k_F3) =
+	@Gtk2::Gdk::Keysyms{qw/Tab ISO_Left_Tab Return KP_Enter BackSpace Escape KP_Multiply Home F3/};
 #my @k_parse_word = ($k_tab, map ord($_), ' ', qw/. ; , ' "/);
 
-our %UNDO_STEPS = (
-	delete     => 'insert',
-	insert     => 'delete',
-	apply_tag  => 'remove_tag',
-	remove_tag => 'apply_tag'
-);
-
 $Gtk2::Ex::HyperTextBuffer::TAGS{link} = [foreground => 'blue'];
 $Gtk2::Ex::HyperTextBuffer::TAGS{underline} = [background => 'yellow'];
 
@@ -209,6 +196,7 @@ sub init { # called by new()
 
 	my $vbox = Gtk2::VBox->new(0, 0);
 	$self->{vbox} = $vbox;
+	$self->{top_widget} = $vbox;
 	
 	my $scrolled_window = Gtk2::ScrolledWindow->new();
 	$scrolled_window->set_policy('automatic', 'automatic');
@@ -229,7 +217,13 @@ sub init { # called by new()
 	$htext->signal_connect(link_clicked =>
 		sub { $self->{app}->link_clicked($_[1]) }  );
 	$htext->signal_connect(link_enter =>
-		sub { $self->{app}->push_status("Go to \"$_[1]\"", 'link') }  );
+		sub {
+			my $p = $self->{app}{page};
+			my ($t, $l) = $p->parse_link($_[1]);
+			#warn "parsed: $_[1] => $l ($t)\n";
+			$l = $p->resolve_name($l) if $t eq 'page';
+			$self->{app}->push_status("Go to \"$l\"", 'link');
+		}  );
 	$htext->signal_connect(link_leave =>
 		sub { $self->{app}->pop_status('link') }  );
 	$htext->signal_connect_after(toggle_overwrite => \&on_toggle_overwrite, $self);
@@ -244,17 +238,21 @@ sub init { # called by new()
 	}
 	else { # 2.4
 		$tlist->add_table(
-			['text/uri-list', [], 0], map [$_, [], 1],
-			qw{UTF8_STRING TEXT COMPOUND_TEXT text/plain} );
+			['text/uri-list', [], 0],
+			$self->list_text_targets(1) );
 	}
+	$tlist->add_table(['text/x-zim-page-list', [], 2]);
 	$htext->drag_dest_set_target_list($tlist);
 	$htext->signal_connect_swapped(
-		drag_data_received => \&on_drag_data_recieved, $self );
+		drag_data_received => \&on_drag_data_received, $self );
 	#$htext->signal_connect(drag_motion => \&on_drag_motion); # debug
 	$scrolled_window->add($htext);
 	$self->{htext} = $htext;
 	$self->{widget} = $htext;
 
+	$self->set_font($self->{app}{settings}{textfont})
+		if defined $self->{app}{settings}{textfont};
+
 	# init search box
 	my $hbox = Gtk2::HBox->new(0, 5);
 	$hbox->set_no_show_all(1);
@@ -267,7 +265,8 @@ sub init { # called by new()
 	$close_button->add(Gtk2::Alignment->new(0.5, 0.5, 0, 0));
 	$close_button->child->add(
 		Gtk2::Image->new_from_stock('gtk-close', 'small-toolbar') );
-	$close_button->signal_connect(clicked => sub { $self->hide_find } );
+	$close_button->signal_connect_swapped(
+		clicked => \&CloseFind, $self );
 	$hbox->pack_end($close_button, 0,1,0);
 	
 	$hbox->pack_start( Gtk2::Label->new(' Find: '), 0,1,0);
@@ -278,15 +277,15 @@ sub init { # called by new()
 	$hbox->pack_start($entry, 0, 1, 0);
 	$self->{entry} = $entry;
 
-	my $prev_button = $self->{app}->new_button('gtk-go-back', '_Previous');
+	my $prev_button = $self->new_button('gtk-go-back', '_Previous');
 	$prev_button->set_sensitive(0);
-	$prev_button->signal_connect_swapped(clicked => \&find_prev, $self);
+	$prev_button->signal_connect_swapped(clicked => \&FindPrev, $self);
 	$hbox->pack_start($prev_button, 0, 1, 0);
 	$self->{find_prev_button} = $prev_button;
 	
-	my $next_button = $self->{app}->new_button('gtk-go-forward', '_Next');
+	my $next_button = $self->new_button('gtk-go-forward', '_Next');
 	$next_button->set_sensitive(0);
-	$next_button->signal_connect_swapped(clicked => \&find_next, $self);
+	$next_button->signal_connect_swapped(clicked => \&FindNext, $self);
 	$hbox->pack_start($next_button, 0, 1, 0);
 	$self->{find_next_button} = $next_button;
 
@@ -294,19 +293,18 @@ sub init { # called by new()
 	$self->add_actions($ui_actions_ro);
 	$self->add_ui($ui_layout_ro);
 	
-	my $read_only = $self->{app}{settings}{read_only};
 	my $accels = $self->{app}{ui}->get_accel_group;
-	unless ($read_only) {
+	unless ($self->{app}{settings}{read_only}) {
 		$self->add_actions($ui_format_actions, undef, 'ApplyFormat');
 		$self->add_actions($ui_format_toggle_actions, 'TOGGLE', 'ApplyFormat');
 		$self->add_actions($ui_copy_paste_actions, undef, 'clipboard');
 		$self->add_actions($ui_actions);
 		$self->add_ui($ui_layout);
 		$accels->connect( # ^Y (identical with the shift-^Z defined above)
-			ord('Y'), ['control-mask'], ['visible'], sub {$self->redo} );
+			ord('Y'), ['control-mask'], ['visible'], sub {$self->Redo} );
 	}
 	$accels->connect( # alt-/ (identical with the ^F defined above)
-		ord('/'), ['mod1-mask'], ['visible'], sub {$self->show_find} );
+		ord('/'), ['mod1-mask'], ['visible'], sub {$self->Find} );
 	$accels->connect( $k_F3, [], ['visible'], sub { $self->on_FindNext } );
 	$accels->connect( $k_F3, ['shift-mask'], ['visible'], sub { $self->on_FindPrev } );
 
@@ -317,11 +315,11 @@ sub init { # called by new()
 	$htext->signal_connect(
 		focus_out_event => sub { $actions->set_sensitive(0) });
 	
-	$htext->set_cursor_visible( $read_only ? 0 : 1 );
-
 	my $sendto_dir = File::Spec->catdir(xdg_data_home(), 'zim', 'SendTo');
 	$self->{sendto_dir} = $sendto_dir if -d $sendto_dir;
 	#warn "sendTo dir: $sendto_dir\n";
+	
+	$self->on_show_cursor;
 }
 
 =back
@@ -450,7 +448,10 @@ sub Link {
 	$link = $text unless defined $link;
 	return if $link =~ /\n/;
 	
-	unless ($self->{app}{settings}{read_only}) {
+	unless (
+		$self->{app}{settings}{read_only} or
+		$self->{app}{page}{properties}{read_only}
+	) {
 		my $bit = $link eq $text;
 		$buffer->remove_all_tags($start, $end);
 		$self->{htext}->apply_link(($bit ? undef : $link), $start, $end);
@@ -499,63 +500,20 @@ sub MergePage {
 
 Undo one editing step in the buffer.
 
-=cut
-
-sub Undo {
-	my $self = shift;
-	my ($undo, $redo) = @{$self}{'undo', 'redo'};
-	return unless @$undo;
-	my $step = pop @$undo;
-	unshift @$redo, [@$step]; # force copy;
-	$$step[0] = $UNDO_STEPS{$$step[0]};
-	$self->_do_step(@$step);
-}
-
 =item C<Redo()>
 
 Redo one editing step in the buffer.
 
 =cut
 
-sub Redo {
+sub Undo {
 	my $self = shift;
-	my ($undo, $redo) = @{$self}{'undo', 'redo'};
-	return unless @$redo;
-	my $step = shift @$redo;
-	push @$undo, $step;
-	$self->_do_step(@$step);
+	$self->{htext}->get_buffer->undo;
 }
 
-sub _do_step {
-	my ($self, $action, $offset, @data) = @_;
-	my $buffer = $self->{buffer};
-	my $start = $buffer->get_iter_at_offset($offset);
-	$self->{undo_lock} = 1;
-	if ($action eq 'insert') {
-		$buffer->insert($start, $data[0]);
-		$buffer->place_cursor(
-			$buffer->get_iter_at_offset($offset + length($data[0])));
-	}
-	elsif ($action eq 'delete') {
-		my $end = $buffer->get_iter_at_offset($offset + length($data[0]));
-		$buffer->delete($start, $end);
-		$buffer->place_cursor($start);
-	}
-	elsif ($action eq 'apply_tag') {
-		my $end = $buffer->get_iter_at_offset( $data[0] );
-		$buffer->remove_all_tags($start, $end);
-		if ($data[1] eq 'link') {
-			$self->{htext}->apply_link($data[2], $start, $end);
-		}
-		else { $buffer->apply_tag_by_name($data[1], $start, $end) }
-	}
-	elsif ($action eq 'remove_tag') {
-		my $end = $buffer->get_iter_at_offset( $data[0] );
-		$buffer->remove_all_tags($start, $end);
-	}
-	$buffer->set_modified(1);
-	$self->{htext}->scroll_mark_onscreen( $buffer->get_insert );
-	$self->{undo_lock} = 0;
+sub Redo {
+	my $self = shift;
+	$self->{htext}->get_buffer->redo;
 }
 
 sub on_clipboard {
@@ -578,78 +536,54 @@ sub Delete {
 	$self->{htext}->signal_emit('delete_from_cursor', 'chars', 1);
 }
 
-sub on_EditLink { pop->edit_link_dialog }
+=item C<EditLink()>
 
-sub on_Find { pop->show_find }
-
-sub on_FindNext { pop->find_next }
-
-sub on_FindPrev { pop->find_prev }
-
-sub on_FindReplace {
-	pop->{app}->FindReplaceDialog->show();
-}
-
-sub on_InsertDate { # insert date that is linked
-	my $self = pop;
-	my $settings = $self->{app}{settings};
-	my $buffer = $self->get_buffer;
-	my $string = Encode::decode($CODESET,
-		strftime( $settings->{date_string}, localtime) );
-	my $link = $settings->{cal_namespace} . strftime('%Y_%m_%d', localtime);
-	$buffer->insert_blocks_at_cursor(['link', {to => $link}, $string]);
-}
-
-sub on_InsertLink { pop->edit_link_dialog }
-
-=item C<InsertImage(IMAGE, ITER)>
-
-Inserts an image into the buffer. Without argument prompts the user for a file.
-Without ITER the cursor position is used.
+Shows the "Edit link" dialog for the link under the cursor.
 
 =cut
 
-sub InsertImage {
-	my ($self, $file, $iter) = @_;
-	$file = $self->{app}->filechooser_dialog unless $file;
-	$file = $self->{app}{repository}->relative_path($file, $self->{app}{page});
-	return unless length $file;
-	$self->{buffer}->insert_image_from_file($file, $iter);
-}
+sub EditLink { shift->edit_link_dialog }
 
+=item C<Find(STRING)>
 
-=item C<widget()>
+Shows the "Find" bar below the TextView.
+STRING is optional query string, which is executed.
 
-Returns the root widget. This should be used to add the object to a container widget.
-Also use this widget for things like show_all() and hide_all().
+=item C<FindNext()>
+
+Repeat previous find in forward direction.
+
+=item C<FindPrev()>
+
+Repeat previous find in backward direction.
+
+=item C<CloseFind()>
+
+Hides the "Find" bar.
+
+=item C<FindReplace(STRING)>
+
+Show the "Find and Replace" dialog.
+STRING is the optional query string.
 
 =cut
 
-sub widget { return $_[0]->{vbox} }
-
-sub on_key_press_event_hbox {
-	my ($hbox, $event, $self) = @_;
-	return 0 unless $event->keyval == $k_escape;
-	$self->hide_find;
-	return 1;
+sub Find {
+	my ($self, $query) = @_;
+	unless ($self->{entry}->visible) {
+		$self->{hbox}->set_no_show_all(0);
+		$self->{hbox}->show_all;
+		$self->{entry}->grab_focus;
+	}
+	if (length $query) {
+		$self->{entry}->set_text($query);
+		$self->_find($query, 0);
+	}
 }
 
-sub on_activate_entry {
-	my ($self, $entry) = @_;
-	$self->{hbox}->hide;
-	$self->_find($entry->get_text, 0);
-	$self->{htext}->grab_focus;
-}
+sub FindNext { $_[0]->_find( $_[0]->{entry}->get_text, 1  ) }
 
-=item C<find(STRING)>
-
-=item C<find_next()>
-
-=item C<find_prev()>
-
-Search for strings in the buffer.
-
-=cut
+sub FindPrev { $_[0]->_find( $_[0]->{entry}->get_text, -1 ) }
 
 sub _find {
 	my $self = shift;
@@ -659,41 +593,165 @@ sub _find {
 		for @{$self}{'find_prev_button', 'find_next_button'};
 }
 
-sub find {
-	$_[0]->{entry}->set_text($_[1]);
-	$_[0]->_find($_[1], 0);
-}
-
-sub find_next { $_[0]->_find( $_[0]->{entry}->get_text, 1  ) }
-
-sub find_prev { $_[0]->_find( $_[0]->{entry}->get_text, -1 ) }
-
-=item C<show_find()>
-
-Show the search bar at the bottom of the page.
-
-=cut
-
-sub show_find {
-	my $self = shift;
-	$self->{hbox}->set_no_show_all(0);
-	$self->{hbox}->show_all;
-	$self->{entry}->grab_focus;
-}
-
-=item C<hide_find()>
-
-Hide the search bar.
-
-=cut
-
-sub hide_find {
+sub CloseFind {
 	my $self = shift;
 	$self->{hbox}->hide_all;
 	$self->{hbox}->set_no_show_all(1);
 	$self->{htext}->grab_focus;
 }
 
+sub FindReplace {
+	$_[0]->{app}->FindReplaceDialog->show($_[1]);
+}
+
+=item C<InsertDate(DATE, NO_ASK)>
+
+Inserts a date into the text.
+DATE is optional and should be an array ref to the fields
+as produced by C<localtime()>.
+By default the current date is used.
+
+A dialog with formatting options is presented to the user
+unless NO_ASK is true.
+
+=item C<prompt_date_format(DATE)>
+
+Used by C<InsertDate()>.
+
+=cut
+
+sub InsertDate { # insert date that is linked
+	my ($self, $date, $no_ask) = @_;
+	$date ||= [localtime];
+	
+	my ($l, $fmt) = $no_ask
+		? (1, $self->{app}{settings}{date_format})
+		: $self->prompt_date_format($date) ;
+	return unless length $fmt;
+	#warn "Got date fmt: $fmt link: $l\n";
+
+	my $buffer = $self->get_buffer;
+	my $string = Encode::decode($CODESET, strftime($fmt, @$date));
+	my $link = $self->{app}{settings}{cal_namespace}
+		. strftime('%Y_%m_%d', @$date) ;
+	
+	$buffer->insert_blocks_at_cursor(
+		($l ? ['link', {to => $link}, $string] : $string), ' ' );
+}
+
+sub prompt_date_format {
+	my ($self, $date) = @_;
+	
+	my $dialog = Gtk2::Dialog->new(
+		'Insert Date - Zim', $self->{app}{window},
+		[qw/modal destroy-with-parent no-separator/],
+		'gtk-cancel'  => 'cancel',
+	);
+	my $button = $self->new_button('gtk-ok', '_Insert');
+	$dialog->add_action_widget($button, 'ok');
+	
+	my $view = Gtk2::SimpleList->new('Date' => 'scalar');
+	$view->set_headers_visible(0);
+	$view->get_selection->set_mode('browse');
+	$view->signal_connect(
+		row_activated => sub { $dialog->response('ok') });
+	$dialog->vbox->add($view);
+
+	my $check = Gtk2::CheckButton->new('_Link to date');
+	my $l = $self->{app}{settings}{recent_date_link};
+	$check->set_active($l);
+	$dialog->vbox->add($check);
+
+	my ($file) = xdg_data_files('zim', 'dates.list');
+	my @dates;
+	if (defined $file) {
+		for (Zim::File->new($file)->read()) {
+			chomp;
+			! /^\s*#/ and /\S/ or next;
+			push @dates, $_;
+		}
+	}
+	else { @dates = ('%A %d/%m/%Y', '%A %d/%m/%Y %H:%M', '%c') }
+	@{$view->{data}} =
+		map Encode::decode($CODESET, strftime($_, @$date)),
+		@dates ;
+	
+	my $i = $self->{app}{settings}{recent_date_format};
+	$view->select($i) if defined $i;
+	
+	$dialog->show_all;
+	my $r = $dialog->run;
+	if ($r eq 'ok') {
+		($i) = $view->get_selected_indices;
+		$self->{app}{settings}{recent_date_format} = $i;
+		$l = $check->get_active;
+		$self->{app}{settings}{recent_date_link} = $l;
+		$dialog->destroy;
+		return ($l, $dates[$i]);
+	}
+	else {
+		$dialog->destroy;
+		return undef;
+	}
+}
+
+=item C<InsertLink(LINK, TEXT, NO_ASK, NL)>
+
+Insert LINK or prompt the user for a link.
+
+NO_ASK is a boolean to suppress the prompt.
+
+=cut
+
+sub InsertLink { 
+	my ($self, $link, $text, $no_ask) = @_;
+	$text = $link unless length $text;
+	if (defined $link and $no_ask) {
+		$self->{buffer}->insert_blocks_at_cursor(
+			['link', {to => $link}, $text], ' ' );
+	}
+	else { $self->edit_link_dialog(undef, $link, $text) }
+}
+
+=item C<InsertImage(IMAGE, NO_ASK)>
+
+Inserts an image into the buffer.
+Without argument prompts the user for a file.
+
+Unless NO_ASK is set it gives a dialog to confirm the action
+which allows to set any properties for the image.
+
+=cut
+
+sub InsertImage {
+	my ($self, $file, $no_ask) = @_;
+	$file = $self->filechooser_dialog unless $file;
+	$file = $self->{app}{page}->relative_path($file);
+	return unless length $file;
+
+	if ($no_ask) {
+		$self->{buffer}->insert_blocks_at_cursor(
+			['image', {src => $file}] );
+	}
+	else {
+		$self->edit_image_dialog(undef, $file);
+	}
+}
+
+sub on_key_press_event_hbox {
+	my ($hbox, $event, $self) = @_;
+	return 0 unless $event->keyval == $k_escape;
+	$self->CloseFind;
+	return 1;
+}
+
+sub on_activate_entry {
+	my ($self, $entry) = @_;
+	$self->{hbox}->hide;
+	$self->_find($entry->get_text, 0);
+	$self->{htext}->grab_focus;
+}
+
 =item C<WordCount()>
 
 Pop up a dialog showing word, line and char count.
@@ -756,6 +814,13 @@ sub WordCount {
 
 =cut
 
+sub on_show_cursor {
+	my $self = shift;
+	$self->{htext}->set_cursor_visible(
+		! $self->{app}{settings}{read_only}
+			|| $self->{app}{settings}{ro_cursor} );
+}
+
 sub on_edit_mode_changed {
 	my ($buffer, $self) = @_;
 	my %tags;
@@ -763,6 +828,18 @@ sub on_edit_mode_changed {
 		my $name = $_->{is_link} ? 'link' : $_->get_property('name');
 		$tags{$name} = 1;
 	}
+
+	if ( $tags{link} ) {
+		my $data = $self->{htext}->get_link_at_cursor;
+		if (defined $data) {
+			my $p = $self->{app}{page};
+			my ($t, $l) = $p->parse_link($data);
+			$l = $p->resolve_name($l) if $t eq 'page';
+			$self->{app}->push_status("Go to \"$l\"", 'link');
+		}
+	}
+	else { $self->{app}->pop_status('link') }
+
 	$self->actions_show_active(
 		TLink      => $tags{link},
 		TBold      => $tags{bold},
@@ -787,37 +864,51 @@ sub on_drag_motion { # debugging purposes only
 	warn "Drag motion with targets: @targets\n";
 }
 
-sub on_drag_data_recieved {
+sub on_drag_data_received {
 	my ($self, $context, $x, $y, $selection, $info, $time) = @_;
 	my $type = $selection->target->name;
 	my $data = $selection->data;
 	#warn "$info ($time): $type: $data\n";
-	return unless $type eq 'text/uri-list';
+
+	return unless grep {$_ eq $type}
+		qw#text/x-zim-page-list text/uri-list#;
 	
-	my @files = grep defined($_), split /[\r\n]+/, $data;
-	for (@files) {
-		s!\%([A-Fa-z0-9]{2})!chr(hex($1))!eg; # url encoding
-		eval {$_ = Encode::decode_utf8($_, 1)}; # utf8 decoding
-		$_ = $self->{app}{repository}->relative_path($_, $self->{app}{page}) if m#^file:/#;
+	my @items = $self->decode_uri_list($data);
+	my @blocks;
+	if ($type eq 'text/x-zim-page-list') {
+		my $rep = $$self{app}{name};
+		@blocks = map {
+			if (s/^\Q$rep\E\?//) {
+				$_ = $self->{app}{page}->relative_name($_);
+			}
+			['link', {to => $_}, $_], "\n"
+		} @items;
+	}
+	else { # 'text/uri-list'
+		@blocks = map {
+			$_ = $self->{app}{page}->relative_path(
+					Zim::File->parse_uri($_) ) 
+				if m#^file:/#;
+			if (m#^\w[\w\+\-\.]+:/#) { # uri
+				['link', {to => $_}, $_], "\n";
+			}
+			else { # path
+				my $mtype = mimetype($_);
+				($mtype =~ /^image\//)
+					? ['image', {src => $_}, $_]
+					: ['link',  {to  => $_}, $_]
+					, "\n" ;
+			}
+		} @items;
 	}
-	@files = map {
-		unless (m#^\w[\w\+\-\.]+:/#) {
-			my $mtype = mimetype($_);
-			($mtype =~ /^image\//)
-				? ['image', {src => $_}, $_]
-				: ['link',  {to  => $_}, $_]
-				, "\n" ;
-		}
-		else { ['link', {to => $_}, $_], "\n" }
-	} @files;
 	
 	($x, $y) = $self->{htext}->window_to_buffer_coords('widget', $x, $y);
 	my $iter = $self->{htext}->get_iter_at_location($x, $y);
-	$self->{buffer}->insert_blocks($iter, @files);
+	$self->{buffer}->insert_blocks($iter, @blocks);
 }
 
 sub on_populate_popup {
-       	# add items to context menu
+	# add items to context menu
 	# object can be a link to a page, a URL (mail / file / http) or an image (which is a file)
 	my ($htext, $menu, $self) = @_;
 	my $iter = $htext->get_iter_at_pointer;
@@ -921,7 +1012,7 @@ sub _sendto_menu {
 		for my $script (Zim::File->list_dir($dir)) {
 			my $item = Gtk2::MenuItem->new_with_label($script);
 			$item->signal_connect(
-				activate => sub {$self->{app}->_exec("$dir/$script", $file)} );
+				activate => sub {$self->_exec("$dir/$script", $file)} );
 			$st_menu->append($item);
 		}
 	}
@@ -948,7 +1039,7 @@ sub on_key_press_event { # some extra keybindings
 	if ($self->{app}{settings}{read_only}) {
 		# Unix like key bindings for read-only mode
 		if ($val == ord '/') {
-			$self->show_find;
+			$self->Find;
 			return 1;
 		}
 		elsif ($val == ord ' ') {
@@ -970,8 +1061,9 @@ sub on_key_press_event { # some extra keybindings
 		return 1;
 	}
 	elsif (
-		$self->{app}{settings}{backsp_unindent} and $val == $k_backspace or
-		$val == $k_tab and $event->state >= 'shift-mask'
+		$self->{app}{settings}{backsp_unindent} and $val == $k_backspace
+		or $val == $k_l_tab
+		or $val == $k_tab and $event->state >= 'shift-mask'
 	) { # BackSpace or Shift-Tab
 		my $buffer = $htext->get_buffer;
 		my ($start, $end) = $buffer->get_selection_bounds;
@@ -1040,18 +1132,21 @@ sub on_key_press_event { # some extra keybindings
 =item C<set_font(STRING)>
 
 Set the default font for the pageview.
+Calling with C<undef> as argument will reset to the default.
 
 =cut
 
 sub set_font {
 	my ($self, $string) = @_;
-	my $font = Gtk2::Pango::FontDescription->from_string($string);
+	my $font = defined($string)
+		? Gtk2::Pango::FontDescription->from_string($string)
+		: undef ;
 	$self->{htext}->modify_font($font);
 }
 
 =item C<get_state()>
 
-Returns a number of properties that need to be saved in the history.
+Returns a HASH ref with properties that need to be saved in the history.
 
 =cut
 
@@ -1059,48 +1154,83 @@ sub get_state {
 	my $self = shift;
 	my $buffer = $self->{buffer} || return;
 	my $cursor = $buffer->get_iter_at_mark($buffer->get_insert)->get_offset;
+	#warn "get_state cursor @ $cursor\n";
 	#my $vscroll = $self->{scrolled_window}->get_vadjustment->get_value;
 	#my $hscroll = $self->{scrolled_window}->get_hadjustment->get_value;
-	return cursor => $cursor, undo => $self->{undo}, redo => $self->{redo};
+	my %rec = (cursor => $cursor);
 		#, vscroll => $vscroll, hscroll => $hscroll;
+	return \%rec;
 }
 
-=item C<set_state(property => value, ..)>
+=item C<set_state(\%PROPETIES)>
 
-Set a number of properties that could be saved in the history.
+Set a number of properties that were saved in the history.
 
 =cut
 
 sub set_state {
 	my $self = shift;
-	my %rec = @_;
+	my %rec = %{shift()};
 	
+	#warn "set state @ $rec{cursor}\n";
 	if (defined $rec{cursor}) {
 		$self->{buffer}->place_cursor(
 			$self->{buffer}->get_iter_at_offset($rec{cursor}) );
 	
 		$self->{htext}->scroll_mark_onscreen( $self->{buffer}->get_insert );
 	}
-	
 	#$self->{scrolled_window}->get_vadjustment->set_value($rec{vscroll});
 	#$self->{scrolled_window}->get_hadjustment->set_value($rec{hscroll});
-	
-	$self->{undo} = $rec{undo} || [];
-	$self->{redo} = $rec{redo} || [];
 }
 
 =item C<load_page(PAGE)>
 
-Load a new page object into the buffer.
+Load a new page in the page view widgets.
 
 =cut
 
 sub load_page {
 	my ($self, $page) = @_;
+	my $mode = $page->properties->{view};
+	$self->on_exit_mode
+		if $self->{mode} and $mode ne $self->{mode};
+	if ($mode and $mode ne $self->{mode}) { # to special mode
+		my $class = 'Zim::GUI::PageView::'.ucfirst(quotemeta($mode));
+		eval "use $class";
+		unless ($@) {
+			bless $self, $class;
+			$self->on_enter_mode;
+		}
+		else { warn $@ }
+	}
+	elsif ($self->{mode} and ! $mode) { # to default mode
+		bless $self, 'Zim::GUI::PageView';
+	}
+	$self->{mode} = $mode;
+	$self->load_page_contents($page);
+}
+
+sub on_enter_mode {} # stub to be overloaded
+
+sub on_exit_mode {} # stub to be overloaded
+
+=item C<load_page_contents(PAGE, TREE)>
+
+To be overloaded by child classes, use C<load_page()> from the GUI.
+
+Load the contents of PAGE into the buffer.
+This method should die when an error is encountered.
+
+TREE is an optional argument, child classes do not need to support it.
+
+=cut
+
+sub load_page_contents {
+	my ($self, $page, $tree) = @_;
 	my $use_spell = defined $self->{app}{objects}{Spell}; # FIXME ugly internals
+	$tree ||= $page->get_parse_tree(); 
 	
 	# clear the old buffer
-	$self->{undo_lock} = 1;
 	$self->{buffer}->clear if $self->{buffer};
 	$self->{app}->Spell->detach($self->{htext}) if $use_spell;
 	$self->{_prev_buffer} = $self->{buffer}; # FIXME hack to prevent segfaults
@@ -1111,33 +1241,26 @@ sub load_page {
 #	$buffer->create_tag('info', editable => 0); #, foreground => 'red');
 	$self->{buffer} = $buffer;
 	$self->{htext}->set_buffer($buffer);
-	my $tree = $page->get_parse_tree;
-	$buffer->set_parse_tree($tree);
+	$buffer->set_parse_tree( $tree );
 	unless ($self->{app}{settings}{read_only}) {
 		# connect signals _after_ load_parsetree()
-		$buffer->signal_connect(delete_range => \&on_delete_range, $self);
-		$buffer->signal_connect_after(insert_text => \&on_insert_text, $self);
 		$buffer->signal_connect(modified_changed =>
 			sub {$self->{app}->update_status} );
-		#$buffer->signal_connect(apply_tag => \&on_apply_tag, $self);
-		#$buffer->signal_connect(remove_tag => \&on_remove_tag, $self);
 		$buffer->signal_connect(edit_mode_changed => \&on_edit_mode_changed, $self);
 	}
+	$buffer->{_zim_name} = $self->{app}{name}; # FIXME FIXME FIXME
 	$self->{app}->Spell->attach($self->{htext}) if $use_spell;
-	$self->{undo_lock} = 0;
 	
 	if ($page->exists) {
 		$buffer->place_cursor(
 	 	      	$buffer->get_iter_at_offset(0) );
 	}
-	else { # new page, place cursor below headers but above footer
+	else { # new page, place cursor below headers
 		my (undef, $iter) = $buffer->get_bounds;
-#		$iter->backward_to_tag_toggle( $buffer->get_tag_table->lookup('info') );
 		$buffer->place_cursor( $iter );
 	}
-	
-	$self->{undo} = [];
-	$self->{redo} = [];
+	on_edit_mode_changed($buffer, $self)
+		unless $self->{app}{settings}{read_only}; # reset format buttons
 	
 	$self->{htext}->scroll_mark_onscreen( $buffer->get_insert );
 
@@ -1162,18 +1285,62 @@ sub modified {
 	$_[0]->{buffer}->get_modified;
 }
 
-=item C<get_parse_tree(PAGE)>
+=item C<save_page(PAGE)>
 
-Returns the parse tree for the current buffer contents.
+Save current buffer contents to PAGE.
 
 =cut
 
-sub get_parse_tree {
-	my $tree = $_[0]->{buffer}->get_parse_tree;
-#	my @head = splice @$tree, 0, 2;
-#	@$tree = grep {not ref($_) && $$_[0] eq 'info'} @$tree;
-#	unshift @$tree, @head;
-	return $tree;
+sub save_page {
+	my ($self, $page) = @_;
+	my $existed = $page->exists;
+	my $ok = $self->save_page_contents($page);
+	if ($ok) {
+		$self->{app}->signal_emit('page_created', $page->name)
+			unless $existed;
+	}
+	else {
+		$page->delete;
+		$self->{app}->signal_emit('page_deleted', $page->name);
+	}
+}
+
+=item C<save_page_contents(PAGE)>
+
+To be overloaded by child classes, use C<save_page()> from the GUI.
+
+This method saves the buffer contents to PAGE.
+Should return 1 on success or 0 when the buffer was empty.
+Returning 0 results in the page being deleted.
+
+This method should die when an error is encountered.
+
+=cut
+
+sub save_page_contents {
+	my ($self, $page) = @_;
+	my $tree = $self->{buffer}->get_parse_tree;
+	return 0 unless @$tree > 2 ; # check content in buffer
+	$page->set_parse_tree($tree);
+	return 1;
+}
+
+=item C<change_widget(WIDGET)>
+
+Replaces the curren page view widget in the interface with WIDGET.
+Returns the old widget.
+
+=cut
+
+sub change_widget {
+	my ($self, $widget) = @_;
+	my $old = $self->{top_widget};
+	#warn "switching $old for $widget\n";
+	$self->{app}{r_vbox}->remove($old);
+	$self->{app}{r_vbox}->pack_end($widget, 1,1,0);
+	$self->{top_widget} = $widget;
+	$self->{app}{vbox}->set_focus_chain($self->{app}{l_vbox}, $widget);
+	return $old;
 }
 
 =item C<parse_backspace(ITER)>
@@ -1181,7 +1348,7 @@ sub get_parse_tree {
 This method is called when the user types a backspace.
 It tries to update the formatting of the current line.
  
-When TRUE is returned the widget does not recieve the backspace.
+When TRUE is returned the widget does not receive the backspace.
 
 =cut
 
@@ -1195,6 +1362,12 @@ sub parse_backspace {
 		$buffer->insert($lf, $line);
 		return 1;
 	}
+	elsif ($line =~ /\t$/) {
+		my $back = $iter->copy;
+		$back->backward_char;
+		$buffer->delete($back, $iter);
+		return 1;
+	}
 	return 0;
 }
 
@@ -1204,7 +1377,7 @@ This method is called when the user is about to insert a linebreak.
 It checks the line left of the cursor of any markup that needs 
 updating. It also takes care of autoindenting.
 
-When TRUE is returned the widget does not recieve the linebreak.
+When TRUE is returned the widget does not receive the linebreak.
 
 =cut
 
@@ -1296,14 +1469,25 @@ sub parse_word {
 	}
 	elsif (
 		$line =~ qr{(?<!\S)(\w[\w\+\-\.]+://\S+)$} # url
-		or $line =~ qr{(?<!\S)([\w\.\-\(\):]*:[\w\.\-\(\)]{2,}:*|\.\w[\w\.\-\(\):]*)$} # page
+		or $line =~ qr{ (?<!\S)(
+		        [\w\.\-\(\)]*(?: :[\w\.\-\(\)]{2,} )+:?
+		      | \.\w[\w\.\-\(\)]+(?: :[\w\.\-\(\)]{2,} )*:?
+		            )$  }x # page (ns:page .subpage)
+		or $line =~ qr{(?<!\S)(
+		      \w[\w\+\-\.]+\?\S+
+		            )$}x # interwiki (name?page)
 		or ( $self->{app}{settings}{use_linkfiles}
-		  and $line =~ qr{(?<!\S)((?:~/[^/\s]|~[^/\s]*/|\.\.?/|/[^/\s])\S*)$} ) # file
+		  and $line =~ qr{ (?<!\S)( (?:
+	  	      ~/[^/\s] | ~[^/\s]*/ | \.\.?/ | /[^/\s]
+		               ) \S* )$  }x ) # file (~/ ~name/ ../ ./ /)
 		or ( $self->{app}{settings}{use_camelcase}
-		  and $line =~ qr{(?<!\S)([[:upper:]]+[[:lower:]]+[[:upper:]]+\w*)$}  ) # CamelCase
+		  and $line =~ qr{(?<!\S)(
+	  	      [[:upper:]]+[[:lower:]]+[[:upper:]]+\w*
+		               )$}x  ) # CamelCase
 	) { # any kind of link
 		my $word = $1;
-		return 0 unless $word =~ /[^:\dhu]/i; # do not link "10:20h"
+		return 0 if $word !~ /[[:alpha:]]{2}/; # at least two letters in there
+		#return 0 unless $word =~ /[^:\dhu]/i; # do not link "10:20h"
 		my ($start, $end) = ($iter->copy, $iter->copy);
 		$start->backward_chars(length $word);
 		return 0 if grep {$_->get_property('name') !~ /spell/}
@@ -1318,7 +1502,8 @@ sub parse_word {
 		
 	}
 	elsif ( $self->{app}{settings}{use_utf8_ent} &&
-		$line =~ /(?<!\S)\\(\w+)$/) { # utf8 chars
+		$line =~ /(?<!\S)\\(\w+)$/
+	) { # utf8 chars
 		my $word = $1;
 		my $chr = _entity($word);
 		return 0 unless defined $chr;
@@ -1495,50 +1680,6 @@ sub selection_backspace {
 	return 0;
 }
 
-sub on_insert_text { # buffer, iter, string, length, self
-#	(length($string) == 1)
-#		? push(@undo_chars, $string)
-#		: 
-	# length in argument list does not understand utf8	
-	$_[4]->add_undo('insert', $_[1]->get_offset - length($_[2]), $_[2]);
-}
-
-sub on_delete_range { # buffer, begin, end, self
-	#print "delete range\n";
-	my $string = $_[0]->get_text($_[1], $_[2], 0);
-	$_[3]->add_undo('delete', $_[1]->get_offset, $string);
-}
-
-#sub on_apply_tag { pop->_on_change_tag('apply_tag', @_) }
-
-#sub on_remove_tag { pop->_on_change_tag('remove_tag', @_) }
-
-sub _on_change_tag {
-	my ($self, $action, undef, $tag, $start, $end) = @_;
-	my @off = ($start->get_offset, $end->get_offset);
-	if ($tag->{is_link}) {
-		$self->add_undo($action, @off, 'L', $tag->{link_data});
-	}
-	else {
-		$self->add_undo($action, @off, $tag->get_property('name'));
-	}
-}
-
-=item C<add_undo(ACTION, OFFSET, DATA)>
-
-=cut
-
-sub add_undo {
-	my $self = shift;
-	return if $self->{undo_lock}; # prohibit unwanted recursion
-#	flush_undo_chars() if @undo_chars;
-	my ($action, $offset, @data) = @_;
-#	print "do: $action \@$offset: >>@data<<\n";
-	push @{$self->{undo}}, [$action, $offset, @data];
-	shift @{$self->{undo}} if @{$self->{undo}} > $self->{app}{settings}{undo_max};
-	@{$self->{redo}} = ();
-}
-
 =item C<edit_link_dialog(ITER)>
 
 This dialog allows the user to create a link for which
@@ -1548,14 +1689,15 @@ If ITER is undefined the cursor is used.
 
 =cut
 
-sub edit_link_dialog {
-	my ($self, $iter) = @_;
+sub edit_link_dialog { # TODO this routine is also use by InsertLink, which is conflicting behavior
+	my ($self, $iter, $link, $text) = @_;
 	my $buffer = $self->{buffer};
 
 	# handle selections
 	my ($start, $end) = $buffer->auto_selection('link', $iter);
 
-	my ($text, $link) = ('', '');
+	$link = '' unless length $link;
+	$text = $link unless length $text;
 	if (defined $start) {
 		$link = $self->{htext}->get_link_at_iter($start);
 		$text = $self->{buffer}->get_text($start, $end, 0);
@@ -1563,12 +1705,12 @@ sub edit_link_dialog {
 	}
 
 	# Run the dialog
-	my $val = $self->{app}->run_prompt(
+	my $val = $self->run_prompt(
 		(defined($start) ? 'Edit Link' : 'Insert Link'),
 		['txt', 'link'], {
 			txt  => ['Text', 'string', $text],
-			link => ['Links to', 'string', $link],
-		}, $LINK_ICON, '_Link' )
+			link => ['Links to', 'page', $link],
+		}, 'gtk-connect', '_Link' )
 		or return;
 	($text, $link) = @$val;
 	return unless $text =~ /\S/ or $link =~ /\S/;
@@ -1595,31 +1737,134 @@ sub edit_link_dialog {
 	# FIXME redundant code from "apply_link" here
 }
 
-=item C<edit_image_dialog()>
+=item C<edit_image_dialog(ITER, SRC)>
+
+Popup a dialog that allows changing the source and properties
+of an image. Both arguments are optional.
 
 =cut
 
-sub edit_image_dialog { # TODO add 'browse' button
-	my ($self, $iter) = @_;
-	my $pixbuf = $iter->get_pixbuf;
-	return unless $pixbuf and defined $pixbuf->{image_src};
+sub edit_image_dialog {
+	my ($self, $iter, $src) = @_;
+
+	my $pixbuf = $iter ? $iter->get_pixbuf : undef ;
+	$src = $pixbuf->{image_src} if $pixbuf and ! length $src;
+	return if $pixbuf and ! length $src;
 	
-	my $src = $pixbuf->{image_src};
-	my $val = $self->{app}->run_prompt(
+	my ($dialog, $entries) = $self->new_prompt(
 		'Edit Image Source',
 		['src'], {src => ['Source', 'file', $src]},
-		$LINK_ICON, '_Link' )
-		or return;
+		'gtk-connect', '_Link' );
+	
+	my ($w, $h, $ratio) = (0, 0, 1);
+	if ($pixbuf) {
+		($w, $h) = ($pixbuf->get_width, $pixbuf->get_height);
+		$ratio = $w / $h;
+	}
 
-	my ($new_src) = @$val;
-	return unless $new_src =~ /\S/ and $new_src ne $src;
+	my $frame = Gtk2::Expander->new_with_mnemonic('_Properties');
+	$dialog->vbox->add($frame);
+	my $vbox = Gtk2::VBox->new();
+	$vbox->set_border_width(12);
+	$frame->add($vbox);
+
+	my $table = Gtk2::Table->new(2, 2, 0);
+	$table->set_row_spacings(5);
+	$table->set_col_spacings(12);
+	$vbox->add($table);
+
+	my $i = 0;
+	for (qw/Width Height/) {
+		my $label = Gtk2::Label->new($_.':');
+		my $align = Gtk2::Alignment->new(0,0.5, 0,0);
+		$align->add($label);
+		$table->attach_defaults($align, 0,1, $i,$i+1);
+		$i++;
+	}
+
+	my $input_w = Gtk2::SpinButton->new_with_range(1, 4*$w || 1, 1);
+	$input_w->set_value($w);
+	$table->attach_defaults($input_w, 1,2, 0,1);
+	my $input_h = Gtk2::SpinButton->new_with_range(1, 4*$h || 1, 1);
+	$input_h->set_value($h);
+	$table->attach_defaults($input_h, 1,2, 1,2);
+
+	my $reset_button = Gtk2::Button->new('_Reset');
+	$vbox->pack_start($reset_button, 0,0,0);
+
+	my ($scale, $block) = ('', 0);
+	if ($pixbuf and $pixbuf->{resized}) {
+		$scale = $pixbuf->{resized}[0] ? 'w' : 'h';
+	}
+	$input_w->signal_connect(value_changed => sub {
+		return if $block;
+		$scale = 'w';
+		$w = $input_w->get_value;
+		$h = int $w / $ratio;
+		$block = 1;
+		$input_h->set_value($h);
+		$block = 0;
+	} );
+	$input_h->signal_connect(value_changed => sub {
+		return if $block;
+		$scale = 'h';
+		$h = $input_h->get_value;
+		$w = int $ratio * $h;
+		$block = 1;
+		$input_w->set_value($w);
+		$block = 0;
+	} );
+	my $init = sub {
+		my $file = $$entries[0]->get_text;
+		$file = File::Spec->rel2abs($file, $self->{buffer}{base})
+			if $self->{buffer}{base};
+		eval { # dies on my win32 installation
+			(undef, $w, $h) = (-f $file and -r _)
+				? (Gtk2::Gdk::Pixbuf->get_file_info($file))
+				: (0,0) ;
+		};
+		if ($w && $h) {
+			$_->set_sensitive(1)
+				for $reset_button, $input_w, $input_h;
+			$ratio = $w / $h;
+			$block = 1;
+			$input_w->set_range(1, 4*$w);
+			$input_w->set_value($w);
+			$input_h->set_range(1, 4*$h);
+			$input_h->set_value($h);
+			$block = 0;
+			$scale = '';
+		}
+		else {
+			$_->set_sensitive(0)
+				for $reset_button, $input_w, $input_h;
+		}
+	};
+	$reset_button->signal_connect(clicked => $init);
+	$$entries[0]->signal_connect(changed => $init);
+	$init->() unless $pixbuf;
+
+	$dialog->show_all;
+	my $r = $dialog->run;
+	$dialog->destroy;
+	return unless $r eq 'ok';
+
+	my $new_src = $$entries[0]->get_text;
+	return unless $new_src =~ /\S/;
 
 	my $buffer = $self->{buffer};
-	my $end = $iter->copy;
-	$end->forward_char;
-	$buffer->delete($iter, $end);
-	$buffer->place_cursor($iter);
-	$self->{buffer}->insert_image_from_file($new_src);
+	if ($iter and $pixbuf) { # delete image
+		my $end = $iter->copy;
+		$end->forward_char;
+		$buffer->delete($iter, $end);
+		$buffer->place_cursor($iter);
+	}
+	my @arg = $scale
+		? ($new_src, ($scale eq 'w') ? ($w, 0) : (0, $h))
+		: ($new_src)  ;
+	my %m;
+	@m{'src', 'width', 'height'} = @arg;
+	$self->{buffer}->insert_blocks_at_cursor(['image', \%m]);
 }
 
 1;
