@@ -1,9 +1,10 @@
 package Zim::History;
 
 use strict;
-use Storable qw/nstore retrieve/;
+use Zim::File;
+use Zim::Page;
 
-our $VERSION = '0.14';
+our $VERSION = '0.19';
 
 =head1 NAME
 
@@ -28,36 +29,50 @@ all redundancy.
 
 The namespace stack keeps track of the namespaces you opened.
 
-The history is saved in a cache file using L<Storable>.
-
 =head1 METHODS
 
 =over 4
 
-=item C<new(FILE, PAGEVIEW, MAX)>
+=item C<new(FILE, MAX, GUI)>
 
 Constructor. Takes the name of the cache file as an argument
-and reads this cache file. The second argument is the pageview
-object. When a new record is made for the current page the C<get_state()> method is
+and reads this cache file. MAX history records are stored in FILE.
+
+The third argument is the GUI object.
+When a new record is made for the current page the C<get_state()> method is
 called on this object to get state information that needs to be saved.
-Third argument is the maximum number of items in the history.
 
 =cut
 
+# We use two stacks here called "hist" and "recent"
+# both contain the same kind of records (hashes with state data for each page)
+# To keep everything in sync we want only one record per page
+# which is linked by reference one or more times.
+
+
 sub new {
-	my ($class, $file, $view, $max) = @_;
+	my ($class, $file, $max, $gui) = @_;
+	$file = Zim::File->new($file) if defined $file and ! ref $file;
 	my $self = bless {
-		file      => $file,
-		PageView  => $view,
-		max       => $max,
-		point     => 0,
-		hist      => [],
-		recent    => [],
+		file   => $file,
+		max    => $max,
+		GUI    => $gui,
+		point  => 0,
+		hist   => [],
+		recent => [],
 	}, $class;
 	$self->read;
 	return $self;
 }
 
+=item C<set_GUI(GUI)>
+
+Set GUI object.
+
+=cut
+
+sub set_GUI { $_[0]{GUI} = $_[1] }
+
 =item C<read>
 
 Read the cache file.
@@ -65,12 +80,38 @@ Read the cache file.
 =cut
 
 sub read {
+	# FIXME hard-coded property name
+	# FIXME more elegant storage model
 	my $self = shift;
-	return unless -f $self->{file} and -r _;
-	my @cache = @{ retrieve($self->{file}) };
-	my $version = shift @cache;
-	return unless $version == $VERSION;
-	@{$self}{qw/point hist recent/} = @cache;
+	return unless $self->{file} and $self->{file}->exists;
+	my $fh = $self->{file}->open('r') or return;
+	<$fh> =~ /^zim: version $VERSION$/ or return;
+	<$fh> =~ /^point: (\d+)$/ or return;
+	$self->{point} = $1;
+	<$fh> =~ /^hist: (\d+)$/ or return;
+	my $h = $1;
+	my %seen;
+	my $i = 0;
+	while ($i < $h) { # read hist
+		<$fh> =~ /^(\S+)\t(\d+)$/ or last;
+		my $rec = $seen{$1} || {name => $1, state => {cursor => $2}};
+		@$rec{'namespace', 'basename'}
+			= Zim::Page->split_name($$rec{name});
+		push @{$self->{hist}}, $rec;
+		$seen{$$rec{name}} = $rec;
+		$i++;
+	}
+	while (<$fh>) { # read recent
+		/^(\S+)\t(\d+)$/ or last;
+		my $rec = $seen{$1} || {name => $1, state => {cursor => $2}};
+		@$rec{'namespace', 'basename'}
+			= Zim::Page->split_name($$rec{name});
+		push @{$self->{recent}}, $rec;
+		$seen{$$rec{name}} = $rec;
+	}
+	$fh->close;
+	$self->{current} = $self->{hist}[ $self->{point} ];
+	#use Data::Dumper; warn "Seen: ", Dumper \%seen;
 }
 
 =item C<write>
@@ -81,77 +122,87 @@ Write the cache file.
 
 sub write {
 	my $self = shift;
-	$self->get_current; # force object to record conversion
-	nstore([$VERSION, @{$self}{qw/point hist recent/}], $self->{file});
+	$self->_save_state;
+	# FIXME hard-coded property name
+	# FIXME more elegant storage model
+	return unless $self->{file};
+	$self->{file}->write(
+		"zim: version $VERSION\n",
+		"point: $$self{point}\n",
+		"hist: ".scalar(@{$$self{hist}})."\n",
+		map "$$_{name}\t$$_{state}{cursor}\n",
+			@{$$self{hist}}, @{$$self{recent}}
+	);
+}
+
+sub _save_state {
+	$_[0]{current}{state} = $_[0]{GUI}->get_state() if $_[0]{GUI};
 }
 
 =item C<set_current(PAGE)>
 
 Give the page object that is going to be displayed in the 
-PageView. Set a new current before updating the PageView
-so the History object has time to save the state of the PageView
-if necessary.
+GUI. Set a new current before updating the GUI so the
+History object has time to save the state of the GUI.
+Returns the history record of the page.
 
 =cut
 
-sub set_current {
-	my ($self, $page) = @_;
-
-	my $current = $self->get_current;
-	my $name = $page->name;
-	$self->{page} = $page; # set _after_ calling get_current()
-	goto RECENT if $current and $name eq $current->{name}; # directly redundant
-	$current = undef;
+sub set_current { # from_hist is an private arg
+	no warnings;
+	my ($self, $name, $from_hist) = @_;
+	$name = ref $name ? $name->name : $name;
+	$self->_save_state; # even if redundant, sync state
+	return $self->{current}
+		if $self->{current} and $name eq $self->{current}{name};
+		# directly redundant
 
+	# lookup record, or create new record
+	my $rec = $self->get_record($name) || { name => $name };
+	@$rec{'namespace', 'basename'} = Zim::Page->split_name($name);
+	my $prev = $self->{current};
+	$self->{current} = $rec;
+	
 	# update hist stack
 	my $point = $self->{point};
 	my $hist = $self->{hist};
 	if    ($point >= $self->{max}) { shift @$hist }
 	elsif (defined $$hist[$point]) { $point++     }
 	splice @$hist, $point; # clear forw stack
-	$$hist[$point] = undef;
+	$$hist[$point] = $rec;
 	$self->{point} = $point;
-	# note: do not call get_current below this line
 
-	# update recent stack
-	RECENT:
+	$self->_set_recent($from_hist || 0);
+
+	return $rec;
+}
+
+sub _set_recent {
+	# update recent stack to reflect the current page
+	# stuff may have been deleted (fro example because the page was deleted)
+	# while the page still is in the history, so check all possibilities
+	my ($self, $from_hist) = @_;
 	my $recent = $self->{recent};
-	for (@$recent) { return if $_->{name} eq $name } # check redundancy
-	my $rec = $current || { map {$_ => $page->$_} qw/name namespace basename/ };
-	$$hist[$point] = $rec; # force linking of ref
+	my $name = $self->{current}{name};
+	if ($from_hist) { # from back / forw / jump
+		return if grep {$_->{name} eq $name} @$recent;
+		# return if exists
+	}
+	else {
+		@$recent = grep {$_->{name} ne $name} @$recent;
+		# remove redundancy
+	}
 	shift @$recent if $#$recent >= $self->{max};
-	push @$recent, $rec;
-
-	#use Data::Dumper; print 'hist, recent: ', Dumper($hist, $recent);
+	push @$recent, $self->{current};
 }
 
 =item C<get_current()>
 
-Returns the curernt history object. When possible asks the PageView
-objects for the current state information.
+Returns the current history object.
 
 =cut
 
-sub get_current {
-	my $self = shift;
-	my $page = $self->{page};
-	my $rec = $self->{hist}[ $self->{point} ];
-	#print "rec: >>$rec<<, page: >>$page<<\n";
-	return $rec unless $page;
-	
-	unless ($rec) { # new record
-		my $name = $page->name;
-		$rec = $self->_get_record($name) || {
-			name => $name,
-			map {$_ => $page->$_} qw/namespace basename/  };
-		$self->{hist}[ $self->{point} ] = $rec;
-		#use Data::Dumper; print "rec: ", Dumper $rec;
-	}
-
-	%$rec = ( %$rec, $self->{PageView}->get_state() );
-
-	return $rec;
-}
+sub get_current { return $_[0]->{current} }
 
 =item C<delete_recent(PAGE)>
 
@@ -175,10 +226,11 @@ Returns the record for this step or undef on failure.
 sub back {
 	my ($self, $i) = @_;
 	return if $i < 1 or $i > $self->{point};
-	$self->get_current;
+	$self->_save_state;
 	$self->{point} -= $i;
-	$self->{page} = undef;
-	return $self->{hist}[ $self->{point} ];
+	$self->{current} = $self->{hist}[ $self->{point} ];
+	$self->_set_recent(1);
+	return $self->{current};
 }
 
 =item C<forw(INT)>
@@ -192,10 +244,32 @@ sub forw {
 	my ($self, $i) = @_;
 	my $forw = $#{$self->{hist}} - $self->{point};
 	return if $i < 1 or $i > $forw;
-	$self->get_current;
+	$self->_save_state;
 	$self->{point} += $i;
-	$self->{page} = undef;
-	return $self->{hist}[ $self->{point} ];
+	$self->{current} = $self->{hist}[ $self->{point} ];
+	$self->_set_recent(1);
+	return $self->{current};
+}
+
+=item C<jump(NAME)>
+
+Go to page NAME in the history stack.
+
+=cut
+
+sub jump {
+	my ($self, $name) = @_;
+	$name = $name->name if ref $name;
+	$self->_save_state;
+	return $self->{current}
+		if $self->{current} and $self->{current}{name} eq $name;
+
+	my ($i) = grep {$self->{hist}[$_]{name} eq $name} 0 .. $#{$self->{hist}};
+	return $self->set_current($name, 1) unless defined $i;
+	$self->{point} = $i;
+	$self->{current} = $self->{hist}[ $self->{point} ];
+	$self->_set_recent(1);
+	return $self->{current};
 }
 
 =item C<get_history()>
@@ -207,7 +281,6 @@ The integer is the index of the current page in this stack.
 
 sub get_history {
 	my $self = shift;
-	$self->get_current;
 	return $self->{point}, @{$self->{hist}};
 }
 
@@ -220,7 +293,7 @@ The integer is the index of the current page in this stack.
 
 sub get_recent {
 	my $self = shift;
-	my $ref = $self->get_current;
+	my $ref = $self->{current};
 	my $i = 0;
 	for (@{$self->{recent}}) {
 		last if $_ eq $ref;
@@ -237,21 +310,23 @@ history. The result can be used as a namespace stack.
 =cut
 
 sub get_namespace {
-	# FIXME set hist_namespace on set_current
-	# refine logic on updating records by get_current
 	my $self = shift;
-	my $current = $self->get_current;
+	my $current = $self->{current};
 	return unless $current;
 
 	my $namespace = $current->{name};
 	#print STDERR "looking for namespace $namespace";
+	my $l = length $namespace;
 	for (
 		reverse(0 .. $self->{point}-1),
 		reverse($self->{point} .. $#{$self->{hist}})
 	) {
 		my $rec = $self->{hist}[$_];
 		my $name = $$rec{_namespace} || $$rec{name};
-		$namespace = $name if $name =~ /^:*$namespace:/;
+		if (substr($name,0,$l+1) eq $namespace.':') {
+			$namespace = $name;
+			$l = length $namespace;
+		}
 	}
 	#print STDERR " => $namespace\n";
 	$current->{_namespace} = $namespace;
@@ -284,19 +359,14 @@ Returns the history record for a given page object or undef.
 =cut
 
 sub get_record {
+	no warnings;
 	my ($self, $page) = @_;
 	
 	my $name = ref($page) ? $page->name : $page;
 	$name =~ s/:+$//;
 	
-	my $current = $self->get_current;
-	return $current if $current->{name} eq $name;
-
-	return $self->_get_record($name);
-}
-
-sub _get_record { # used by get_current()
-	my ($self, $name) = @_;
+	return $self->{current}
+		if $self->{current}{name} eq $name;
 
 	#print STDERR "get_record $name\n";
 	for (@{$self->{hist}}, @{$self->{recent}}) {
