@@ -6,14 +6,10 @@ use POSIX qw(strftime);
 use Encode;
 use File::Spec;
 use File::MimeInfo;
-use Zim::File;
-use Zim::Repository::Base;
-use Zim::Page::Text;
+use Zim::Repository;
 
-our $VERSION = '0.17';
-our @ISA = qw/Zim::Repository::Base/;
-
-my $case_tolerant_fs = File::Spec->case_tolerant();
+our $VERSION = '0.18';
+our @ISA = qw/Zim::Repository/;
 
 *CODESET = \$Zim::CODESET;
 $CODESET ||= 'utf8';
@@ -31,7 +27,7 @@ See L<Zim::Repository> for the interface documentation.
 
 =over 4
 
-=item C<new(NAMESPACE, DIR)>
+=item C<new(PARENT, NAMESPACE, DIR)>
 
 Simple constructor. DIR is the root directory of the repository.
 NAMESPACE is the namespace that maps to that directory.
@@ -39,18 +35,31 @@ NAMESPACE is the namespace that maps to that directory.
 =cut
 
 sub init { # called by new
-	my ($self, $dir, $format) = @_;
+	my $self = shift;
 	
-	$dir = Zim::File->abs_path($dir);
-	$self->{dir} = $dir;
-	$self->{format} = $format || 'wiki';
-	$self->{ext} = ($self->{format} eq 'wiki') ? 'txt' : 'html';
+	unless (defined $self->{dir}) {
+		# no dir given, find one based on namespace
+		die "Zim::Repository::Files needs a directory to initialize\n"
+			unless length $self->{parent}{dir};
+		my @parts = grep length($_), split /:+/, $self->{namespace};
+		$self->{dir} = Zim::File->resolve_file(
+			{is_dir => 1}, $$self{parent}{dir}, @parts );
+	}
+	else { # check dir relative to parent
+		$self->{dir} = Zim::File->abs_path(
+				$self->{dir}, $self->{parent}{dir} );
+	}
+	$self->{config}{read_only} = (-w $self->{dir}) ? 0 : 1;
+	$self->{cdir} = $self->{dir} . '/_changes';
+	$self->{format} ||= 'wiki';
+	$self->{ext} = ($self->{format} eq 'html')     ? 'html' :
+	               ($self->{format} eq 'txt2tags') ? 't2t'  : 'txt' ;
 		# FIXME HACK FIXME - this belongs in a Formats.pm
 	
-	$self->{cache} = Zim::File->new($dir, '.zim.cache');
+	$self->{cache} = Zim::File->new($self->{dir}, '.zim.cache');
 	
 	# Check version of cache
-	if (-w $self->{cache}->dir) {
+	unless ($self->{config}{read_only}) {
 		my $line = '';
 		if ($self->{cache}->exists) {
 			my $fh = $self->{cache}->open();
@@ -64,10 +73,6 @@ sub init { # called by new
 	return $self;
 }
 
-=item C<has_pages(NAMESPACE)>
-
-Returns boolean whether NAMESPACE exists.
-
 =item C<list_pages(NAMESPACE)>
 
 Returns a list of pages. If possible it uses a cache.
@@ -96,13 +101,6 @@ Returns a list of pages. If possible it uses a cache.
 # a:c: mtime > links
 # a:c:d mtime > links
 
-sub has_pages {
-	my ($self, $namespace) = @_;
-
-	my $dir = $self->dir($namespace);
-	return -d $dir;
-}
-
 sub list_pages {
 	my ($self, $namespace) = @_;
 	
@@ -111,8 +109,9 @@ sub list_pages {
 
 	my $mtime = (stat $dir)[9];
 	my ($cache_mtime, @pages);
-	for ($self->{cache}->read) {
-		/^\Q$namespace\E(?:([^:\s]+:?) \d+ >| (\d+) \/)/ or next;
+	my $re = qr/^\Q$namespace\E(?:([^:\s]+:?) \d+ >| (\d+) \/)/;
+	for ($self->{cache}->grep($re, 'lines')) {
+		$_ =~ $re or next;
 		if (defined $1 and length $1) { push @pages, $1 }
 		else { # namespace itself - check index time
 			$cache_mtime = $2;
@@ -131,7 +130,7 @@ sub _flush_cache {
 	$self->{cache}->remove if $self->{cache}->exists;
 }
 
-sub _cache_dir {
+sub _cache_dir { # FIXME Can this be optimized ??
 	my ($self, $namespace, $dir) = @_;
 	warn "# Indexing $namespace\n";
 	
@@ -139,11 +138,11 @@ sub _cache_dir {
 		grep defined($_),
 		map {
 			my $item = "$dir/$_";
-			s/[^:\w\.\-\(\)]/_/g;
+			s/([^[:alnum:]_\.\-\(\)])/sprintf("%%%02X",ord($1))/eg;
 			(-d $item)		? [$_.':' => $item] :
 			(s/\.$$self{ext}$//)	? [$_ => $item]     : undef ;
 		} 
-		grep /^\w/, Zim::File->list_dir($dir);
+		grep /^[[:alnum:]]/, Zim::File->list_dir($dir);
 	#use Data::Dumper; warn Dumper \@pages;
 	
 	@pages = sort {lc($$a[0]) cmp lc($$b[0])} @pages;
@@ -153,7 +152,7 @@ sub _cache_dir {
 	@pages = grep defined($_), @pages;
 	#use Data::Dumper; warn Dumper \@pages;
 
-	return map {$$_[0]} @pages if $self->{parent}{config}{read_only};
+	return map {$$_[0]} @pages if $self->{config}{read_only};
 	
 	my %items = ();
 	my $index = '';
@@ -189,11 +188,18 @@ sub _cache_page {
 	my ($self, $page) = @_;
 	my $name = $page->name;
 	my ($index, $is_dir);
+	return if $self->{config}{read_only};
 	for ($self->{cache}->read) {
 		if (/^\Q$name\E(:?) \d+ >/) { $is_dir = $1 }
 		else { $index .= $_ }
 	}
-	$self->{cache}->write($index, $self->_cache_string($page, $is_dir))
+	if ($page->{status} eq 'deleted') {
+		$self->{cache}->write($index);
+	}
+	else {
+		$self->{cache}->write($index,
+			$self->_cache_string($page, $is_dir));
+	}
 }
 
 sub _cache_string {
@@ -203,21 +209,30 @@ sub _cache_string {
 		$page = $self->get_page($page);
 	}
 	my $mtime = (stat $page->{source}->path)[9] || '0';
-	my @links = eval{ $page->_list_links };
+	my @links = eval{ $page->list_links };
 	my $key = $page->name;
 	$key .= ':' if $is_dir;
 	return $key . ' ' . $mtime . ' > ' . join(' ', @links) . "\n" ;
 }
 
-sub _list_backlinks {
+=item C<list_backlinks(PAGE)>
+
+Returns a list with names of pages that link to this page.
+
+=cut
+
+sub list_backlinks {
 	my ($self, $page) = @_;
-	my $name = $page->name;
+	my $name = ref($page) ? $page : $page->name;
 	my @links;
-	for ($self->{cache}->read) {
-		/^(:\S+) \d+ >.*?\s\Q$name\E\s/ or next;
+	# Used the regex: qr/^:\S+ \d+ >.*?\s\Q$name\E\s/
+	# here before, but this can become very slow when utf8 is involved
+	# so using two stage grep to make it more scalable
+	for ($self->{cache}->grep(qr/\s\Q$name\E(\s|$)/, 'lines')) {
+		$_ =~ /^(:\S+) \d+ >/ or next;
 		my $l = $1;
 		$l =~ s/:$//;
-		push @links, $l;
+		push @links, $l unless $l eq $name;
 	}
 	return @links;
 }
@@ -225,7 +240,7 @@ sub _list_backlinks {
 sub _search { # query is a hash ref with options etc
 	my ($self, $query, $callback, $ns) = @_;
 	$ns ||= $self->{namespace};
-	warn "# Searching: $ns\n";
+	warn "Searching $ns\n";
 	
 	my $reg = $$query{regex};
 	unless ($reg) {
@@ -240,8 +255,8 @@ sub _search { # query is a hash ref with options etc
 	for ($self->list_pages($ns)) {
 		my $p = $ns.$_;
 		my $is_dir = ($p =~ s/:$//);
-		my $match = $self->file($p)->grep($reg);
-		#warn "$1 matches $match times\n" if $match;
+		my $match = ($p =~ $reg) ? 1 : 0 ;
+		$match += $self->file($p)->grep($reg, 'count');
 		$callback->($match ? [$p, $match] : ());
 		$self->_search($query, $callback, $p.':') if $is_dir; # recurs
 	}
@@ -253,8 +268,9 @@ sub _match_word {
 	$word =~ s/[^\w\.\:\-]/_/g;
 	my $seen = 0;
 	#warn "looking up \"$word\" in $namespace\n";
-	for ($self->{cache}->read) {
-		next unless /^\s*\Q$namespace\E(?i)\Q$word\E(_|:?\s)/;
+	my $re = qr/^\Q$namespace\E(?i)\Q$word\E(_|:?\s)/;
+	for ($self->{cache}->grep($re)) {
+		$_ =~ $re or next;
 		if ($1 eq '_') { return 2 }
 		elsif ($seen) { return 2 }
 		else { $seen = 1 }
@@ -264,14 +280,14 @@ sub _match_word {
 
 =item C<get_page(PAGE_NAME)>
 
-Returns an object of the type L<Zim::Page::Text>.
+Returns an object of the type L<Zim::Page>.
 
 =cut
 
 sub get_page {
 	my ($self, $name, $source) = @_; # source is a private argument
 
-	my $page = Zim::Page::Text->new($self, $name);
+	my $page = Zim::Page->new($self, $name);
 	$source ||= $self->file($name); # case sensitive lookup
 	$page->set_source($source);
 	$page->set_format($self->{format});
@@ -281,49 +297,79 @@ sub get_page {
 		$page->{parse_tree} = $self->_template($page);
 		$page->status('new');
 	}
+	$page->properties->{read_only} = $self->{config}{read_only};
 
 	return $page;
 }
 
-=item C<resolve_page(NAME)>
+=item C<resolve_case(\@LINK, \@PAGE)>
 
-Like C<get_page()> but with case in-sensitive lookup for a page name.
+See L<Zim::Repository>.
 
 =cut
 
-sub resolve_page {
-	my ($self, $name) = @_;
-	my $source = $self->file($name, 1); # case tolerant lookup
-	$name = $self->pagename($source->path);
-	#warn "Resolved page: $name => $file => $page\n";
-	$self->get_page($name, $source);
+sub resolve_case {
+	my ($self, $link, $page) = @_;
+	my $match;
+	if ($page and @$page) {
+		#warn "resolve_case: @$link @ @$page\n";
+		my $anchor = shift @$link;
+		for (reverse  -1 .. $#$page) {
+			my $t = ':'.join(':', @$page[0..$_], $anchor);
+			#warn "\ttrying: $t\n";
+			## FIXME FIXME optimize the two below together
+			my $file = $self->file($t, 1);
+			my $dir = $self->dir($t, 1);
+			next unless -f $file or -d $dir;
+			$match = join ':', $t, @$link;
+			last;
+		}
+	}
+	else { $match = ':' . join ':', @$link } # absolute
+
+	return undef unless $match;
+	my $file = $self->file($match, 1);
+	return $self->pagename($file);
 }
 
+#sub resolve_page {
+#	my ($self, $name) = @_;
+#	my $source = $self->file($name, 1); # case tolerant lookup
+#	#warn "Resolved $name => $source\n";
+#	$name = $self->pagename($source->path);
+#	return defined($name) ? $self->get_page($name, $source) : undef;
+#}
+
 sub _template {
 	# FIXME make template configurable
 	my ($self, $page) = @_;
 	$page->name =~ /([^:]+):*$/;
 	my $title = ucfirst($1);
 	$title =~ s/_/ /g;
-	my $format = $self->root->{date_string} || '%A %d/%m/%Y';
+	my $format = $self->root->{date_format} || '%A %d/%m/%Y';
 	my $date = Encode::decode($CODESET,
 		strftime($format, localtime)   );
-	return	['Document', $page->properties,
+	return ['Page', {%{$page->properties}},
 			['head1', {}, $title],
 			['Para',  {empty_lines => 1}, "Created $date\n"]
-		];
+	];
 }
 
-=item C<copy_page(SOURCE, TARGET)>
+=item C<copy_page(SOURCE, TARGET, UPDATE_LINKS)>
 
 =cut
 
 sub copy_page {
-	my ($self, $old, $new) = @_;
+	my ($self, $old, $new, $update) = @_;
 	my $source = $self->file($old);
 	my $target = $self->file($new);
 	Zim::File->copy($source, $target);
 	@$new{'status', 'parse_tree'} = ('', undef);
+	if ($update) {
+		my ($from, $to) = ($source->name, $target->name);
+		$self->get_page($_)->update_links($from => $to)
+			for $source->list_backlinks ;
+	}
 }
 
 =item C<move_page(SOURCE, TARGET)>
@@ -338,24 +384,14 @@ sub move_page {
 	my $target = $self->file($new);
 
 	die "No such page: $source\n" unless $source->exists;
-	if ($old->equals($new)) { # maybe user wants to change case
-		$target = $target->{case_sensitive}
-			if defined $target->{case_sensitive};
-		$new->{name} = $self->pagename($target);
-		#warn "set name: $new->{name}\n";
-	}
 	#warn "Moving $source to $target\n";
 	Zim::File->move($source, $target);
 
-	# Move tree below file
-	# FIXME make this optional (?) - think boutupdating links here
-	#my $old_tree = $self->dir($old);
-	#my $new_tree = $self->dir($new);
-	#Zim::File->move($old_tree, $new_tree) if -d $old_tree;
-	
 	# update objects
 	@$old{'status', 'parse_tree'} = ('deleted', undef);
 	@$new{'status', 'parse_tree'} = ('', undef);
+	$self->_cache_page($old);
+	$self->_cache_page($new);
 }
 
 =item C<delete_page(PAGE)>
@@ -374,6 +410,7 @@ sub delete_page {
 	}
 	
 	@$page{'status', 'parse_tree'} = ('deleted', undef) if ref $page;
+	$self->_cache_page($page);
 }
 
 =item C<search()>
@@ -416,8 +453,16 @@ sub file {
 		$page = $page->name;
 	}
 
-	$page =~ s/^:*$self->{namespace}:*//i;
-	my @parts = grep length($_), split /:+/, $page;
+	if ($page.':' eq $self->{namespace}) { # index page sub namespace
+		$page = '_index';
+	}
+	else {
+		$page =~ s/^:*\Q$$self{namespace}\E:*//i;
+	}
+
+	my @parts =
+		map {s/\%([A-Fa-z0-9]{2})/chr(hex($1))/eg; $_}
+		grep length($_), split /:+/, $page;
 
 	my $file = $case_tolerant
 		? Zim::File->resolve_file({ext => $$self{ext}}, $$self{dir}, @parts)
@@ -425,6 +470,8 @@ sub file {
 
 	#warn "\t=> $file\n";
 	$file = Zim::File->new($file);
+	$file->set_check_mtime(1);
+#	$file->signal_connect(write => \&_save_changes, $self);
 	return $file;
 }
 
@@ -433,8 +480,15 @@ sub dir {
 
 	if (ref $page) { $page = $page->name } # looking for dir _below_
 
-	$page =~ s/^:*$self->{namespace}:*//i;
-	my @parts = grep length($_), split /:+/, $page;
+	if ($page.':' eq $self->{namespace}) { # index page
+		return $$self{dir};
+	}
+	else {
+		$page =~ s/^:*\Q$self->{namespace}\E:*//i;
+	}
+	my @parts =
+		map {s/\%([A-Fa-z0-9]{2})/chr(hex($1))/eg; $_}
+		grep length($_), split /:+/, $page;
 
 	my $dir = $case_tolerant
 		? Zim::File->resolve_file({ext => $$self{ext}, is_dir => 1}, $$self{dir}, @parts)
@@ -455,40 +509,40 @@ sub pagename {
 	#warn "looking up pagename for: $file\n";
 	$file = File::Spec->abs2rel($file, $self->{dir})
 		if File::Spec->file_name_is_absolute($file);
-	my @parts = grep length($_), File::Spec->splitdir($file);
-	$parts[-1] =~ s/\.$$self{ext}$//;
+	my @parts =
+		map {s/([^[:alnum:]_\.\-\(\)])/sprintf("%%%02X",ord($1))/eg; $_}
+		grep length($_), File::Spec->splitdir($file);
+	return undef unless @parts;
+	$parts[-1] =~ s/\.\Q$$self{ext}\E$//;
 	return $self->{namespace} . join ':', @parts;
 }
 
-=item C<get_source(SOURCE, MODE)>
+=item C<save_changes(FILE)>
 
-Returns an IO::File object or undef.
+Tries to save changes in the "_changes" dir.
+FILE should be a L<Zim::File> object. Returns boolean for succes.
+Needs the 'diff' command to work.
 
 =cut
 
-sub get_source {
-	my ($self, $source, $mode) = @_;
-	
-	my $file = $source->path;
-	my $cb = undef;
-	if ($mode eq 'r') {
-		return undef unless -e $file;
-		die "File not readable: $file\n" unless -r _;
-		$source->{mtime} = $source->stat->{mtime};
-	}
-	elsif ($mode eq 'w') {
-		die "File not writable: $file\n" if -e $file && ! -w _;
-		die "File has changed on disk since reading\n"
-			if  defined $source->{mtime}
-			and $source->{mtime} < $source->stat->{mtime} ;
-		$source->make_dir;
-		$cb = sub { $source->{mtime} = $source->stat->{mtime} };
-	}
-	else { die "unknown mode: $mode" }
-	
-	return $source->open($mode, $cb);
+sub save_changes { _save_changes($_[1], $_[1]->path, $_[0]) }
+
+sub _save_changes {
+	my ($file, $path, $self) = @_;
+	return 0 unless $self->{cdir};
+	#warn "saving changes for $file\n";
+	my $old = File::Spec->abs2rel($path, $self->{dir});
+	$old = File::Spec->catfile($self->{cdir}, $old);
+	Zim::File->touch($old);
+	my $log = Zim::File->new($old.'.log');
+	$log->append("%% $ENV{USER} ".localtime()."\n");
+	`diff $old $file >> $log`;
+	Zim::File->copy($file, $old);
+
+	return 1;
 }
 
+
 1;
 
 __END__
