@@ -2,14 +2,10 @@ package Zim::Repository::Man;
 
 use strict;
 use File::Spec;
-use Zim::Repository::Base;
-use Zim::Page::Text;
-use Zim::File;
+use Zim::Repository;
 
-our $VERSION = '0.15';
-our @ISA     = 'Zim::Repository::Base';
-
-my $null = File::Spec->devnull;
+our $VERSION = '0.18';
+our @ISA     = 'Zim::Repository';
 
 =head1 NAME
 
@@ -30,10 +26,11 @@ It derives from L<Zim::Repository>.
 
 sub init {
 	my $self = shift;
-	open MAN, "man -w 2> $null |";
+	open MAN, "man -w |";
 	my $path = join '', <MAN>;
 	close MAN;
 	$self->{path} = [grep length($_), split /:+/, $path];
+	$self->{no_show_in_sidepane} = 1; # temp HACK
 	#warn "MANPATH: @{$self->{path}}\n";
 }
 
@@ -45,7 +42,7 @@ Lists all manpages in a certain section.
 
 sub list_pages {
 	my ($self, $namespace) = @_;
-	$namespace =~ s/^$self->{namespace}:*//;
+	$namespace =~ s/^\Q$self->{namespace}\E:*//;
 	return $self->list_sections unless length $namespace;
 	return unless $namespace =~ /^(\d+\w*):*$/;
 	my $section = $1;
@@ -80,7 +77,7 @@ sub list_sections {
 
 =item C<get_page(NAME)>
 
-Returns a L<Zim::Page::Text> object for man page NAME.
+Returns a page object for man page NAME.
 
 =cut
 
@@ -88,12 +85,83 @@ sub get_page {
 	my ($self, $name) = @_;
 	#warn "Get man page: $name\n";
 	
-	$name =~ s/^$self->{namespace}:*//;
-	$name =~ s/^(\d+\w*):+//;
+	my $n = $name;
+	$n =~ s/^\Q$self->{namespace}\E:*//;
+	$n =~ s/^(\d+\w*):+//;
 	my $sect = $1 || '';
 	
+	open MAN, "man -w $sect $n |";
+	my $path = join '', <MAN>;
+	close MAN;
+	$path = undef unless $path =~ /\S/;
+	return unless $path;
+
+	my $page = Zim::Man::Page->new($self, $name);
+	$page->{format} = '_man'; # to force formatted interface
+	$page->{_manpage} = [$n, $sect];
+	
+	return $page;
+}
+
+=item C<resolve_name(NAME, REF, EXIST)>
+
+Case in-sensitive check whether a page exists or not.
+REF is ignored since man pages don't have relative links.
+
+=cut
+
+sub resolve_name {
+	my ($self, $name, undef, $exist) = @_;
+	$name = ':'.$name;
+	$name =~ s/^:*\Q$self->{namespace}\E:*(?:(\d+\w*):+)?//i;
+	$name =~ s/^://;
+	my $sect = lc($1) || '';
+	$sect = lc($1) if $name =~ s/\((\d+\w*)\):*$//;
+
+	my @try = ($name);
+	push @try, lc $name if $name =~ /[[:upper:]]/;
+	if ($name =~ /:/) {
+		my $n = $name;
+		$n =~ s/:+/::/; # perl modules / namespaces
+		push @try, $n, lc $n;
+	}
+
+	my ($path, $n);
+	for (@try) {
+		$n = $_;
+		#warn "Resolving $name in section $sect\n";
+		open MAN, "man -w $sect $n |";
+		$path = join '', <MAN>;
+		close MAN;
+		$path = undef unless $path =~ /\S/;
+		last if $path;
+	}
+	$sect = lc($1) if ! $sect and $path =~ /man(\d+\w*)\W/;
+	$name = $self->{namespace}.($sect ? $sect.':' : '').($n || $name);
+
+	return $name if $path;
+	return $exist ? undef : $name ;
+}
+
+package Zim::Man::Page;
+
+use Zim::Page;
+
+our @ISA = qw/Zim::Page/;
+
+=item C<get_parse_tree()>
+
+Returns parse tree for man page.
+
+=cut
+
+sub get_parse_tree {
+	my ($self) = @_;
+	return $self->{parse_tree} if defined $self->{parse_tree};
+	my ($name, $sect) = @{$self->{_manpage}};
+	
 	$ENV{MANWIDTH} = 80; # FIXME get window size (via Env ?)
-	open MAN, "man -c $sect $name 2> $null |" or return undef;
+	open MAN, "man -c $sect $name |" or return undef;
 	my ($block, @data);
 	while (<MAN>) {
 		# FIXME implement parsing algo like in Zim.pm
@@ -119,54 +187,33 @@ sub get_page {
 	}
 	push @data, $block if length $block;
 	close MAN;
-	return undef unless @data;
 	
-	$name = $self->{namespace}.($sect ? $sect.':' : '').$name;
-	my $page = Zim::Page::Text->new($self, $name);
-	$page->set_parse_tree(['Document', {}, @data]);
-	
-	return $page;
+	$self->{parse_tree} = ['Page', {}, @data];
+	return $self->{parse_tree};
 }
 
-=item C<resolve_link(PAGE, LINK)>
+=item C<set_parse_tree(TREE)>
 
-Stub that calls C<resolve_page()>.
-
-Man pages do not contain relative links or links to pages in other namespaces.
-They can only link to other man pages.
+Man pages can not be saved, thus it throws away all data.
 
 =cut
 
-sub resolve_link { $_[0]->resolve_page($_[2]) }
+sub set_parse_tree { warn "Man pages are not writable.\n" }
 
-=item C<resolve_page(NAME)>
+=item C<parse_link(LINK)>
 
-Case in-sensitive check whether a page exists or not.
+Man pages contain no relative links, only absolute in same namespace.
 
 =cut
 
-sub resolve_page {
-	my ($self, $name) = @_;
-	$name = ':'.$name;
-	$name =~ s/^:*$self->{namespace}:*(?:(\d+\w*):+)?//i;
-	$name =~ s/^://;
-	my $sect = lc($1) || '';
-	$sect = lc($1) if $name =~ s/\((\d+\w*)\):*$//;
-
-	#warn "Resolving $name in section $sect\n";
-	open MAN, "man -w $sect $name 2> $null |";
-	my $path = join '', <MAN>;
-	close MAN;
-
-	$path = undef unless $path =~ /\S/;
-	$sect = lc($1) if ! $sect and $path =~ /man(\d+\w*)\W/;
-	
-	$name = $self->{namespace}.($sect ? $sect.':' : '').$name;
-	return $path ? $self->get_page($name)        : ($name =~ /[A-Z]/)
-	             ? $self->resolve_page(lc $name) : undef              ;
-
+sub parse_link {
+	my $self = shift;
+	my ($t, $l) = Zim::Formats->parse_link(@_);
+	$l =~ s/^:*/$self->{namespace}/ if $t eq 'page';
+	return ($t, $l);
 }
 
+
 1;
 
 __END__
